"""
Function that checks whether all provided arguments are non-falsy.

Example:
    is_set(param_x, param_y)
    This will return True if BOTH params are truthy (not None, False, 0, '').
"""
def is_set(*args):
    return all(arg is not None for arg in args)


import warnings

def _get_max_span(obj, span, comparison):

    if not hasattr(comparison, "frequency"):
        return span

    obj_frequency = (
        getattr(obj.settings.span, "frequency", None)
        or getattr(obj.settings, "frequency", None)
    )

    if comparison.frequency is not None and obj_frequency is not None:
        if comparison.frequency != obj_frequency:
            warnings.warn("Multiple frequencies are not allowed.")
            return span

    if span is None:
        span = comparison
        return span

    start = min(span.start, comparison.start)
    end = max(span.end, comparison.end)

    return start >> end

# Method for Table and Chart
def get_span(obj):
    span = None

    def resolve_span_recursive(node, current_span):
        if hasattr(node, "settings") and hasattr(node.settings, "span") and hasattr(node, "data") and hasattr(node.data, "span"):

            comparison = node.settings.span.resolve(node.data.span)
            current_span = _get_max_span(obj, current_span, comparison)

        if hasattr(node, "CHILDREN"):
            for child in node.CHILDREN:
                current_span = resolve_span_recursive(child, current_span)

        return current_span

    span = resolve_span_recursive(obj, span)

    # if hasattr(obj, "settings") and hasattr(obj.settings, "span"):
    if ".start" in str(obj.settings.span.start):
        obj.settings.span = span.start >> obj.settings.span.end

    if ".end" in str(obj.settings.span.end):
        obj.settings.span = obj.settings.span.start >> span.end

    if obj.settings.span.start > obj.settings.span.end:
        raise Exception(
            f"Invalid highlight range for {obj.__class__.__name__}: Start date ({obj.settings.span.start}) is after the end date ({obj.settings.span.end}). "
            f"Please provide a highlight range where start precedes the end. "
            f"Current range: {obj.settings.span.start} >> {obj.settings.span.end}"
        )

    return obj.settings.span

# Method for Table and Chart
def get_highlight(obj):

    if obj.settings.highlight is None:
        return # Just exit the function

    span = get_span(obj)

    if span is None:
        raise Exception(f"{obj.__class__.__name__} span is missing.")

    start = obj.settings.highlight.start

    if ".start" in str(start):
        start = span.start

    if start not in span:
        # We just warn user. But this start not in span is not a problem, because we return min/max
        # It is just for user to know of issue
        warnings.warn(f"{obj.__class__.__name__} highlight start is outside the selected span.", UserWarning)

    end = obj.settings.highlight.end

    if ".end" in str(end):
        end = span.end

    if start > end:
        raise Exception(
            f"Invalid highlight range for {obj.__class__.__name__}: Start date ({start}) is after the end date ({end}). "
            f"Please provide a highlight range where start precedes the end. "
            f"Current range: {start} >> {end}"
        )

    return max(start, span.start), min(end, span.end)


#TODO: extract this function to external module.
from typing import Literal
from rephorm.objects.table_section import TableSection
from rephorm.objects.table_series import TableSeries

# Adjustment for text width, because fpdf does not calculate it correctly
# Todo: Investigate why, and if it is possible to fix it
magic_number = 10

def get_max_numerical_width(table, pdf, font_family, font_size, biggest_nr, table_element=None, max_width = 0):
    """
    Determine the maximum width needed for numerical values in the table.
    Ensures width does not exceed max_width.

    How it works: Finds the longest formatted numerical string within the whole table
    """
    pdf.set_font(font_family, size=font_size)
    max_cap = pdf.get_string_width(str(biggest_nr)) + magic_number

    table_element = table_element if table_element else table

    for child in table_element.CHILDREN:
        if isinstance(child, TableSeries):
            series_data = child.data.get_values(child.settings.span)
            for value in series_data:
                formatted_value = f"{value:.{child.settings.decimal_precision}f}"
                text_width = pdf.get_string_width(str(formatted_value)) + magic_number
                if text_width > max_width:
                    max_width = min(text_width, max_cap)

        elif isinstance(child, TableSection):
            section_max_width = get_max_numerical_width(table, pdf, font_family, font_size, biggest_nr, child, max_width)
            max_width = max(max_width, section_max_width)

    return max_width

def get_max_width(table, pdf, font_family, font_size, table_element = None, parameter: Literal["title", "unit"] = "unit", max_width = 0):

    pdf.set_font(font_family, size=font_size)

    table_element = table_element if table_element else table

    for child in table_element.CHILDREN:

        if isinstance(child, TableSeries):
            attr_value = getattr(child, parameter)
            if attr_value is not None:
                text_width = pdf.get_string_width(attr_value) + magic_number
                if text_width > max_width:
                    max_width = text_width

        elif isinstance(child, TableSection):
            if parameter == "title":
                section_title_width = pdf.get_string_width(child.TITLE) + magic_number
                if section_title_width > max_width:
                    max_width = section_title_width
            section_max_width = get_max_width(table, pdf, font_family, font_size, child, parameter)
            max_width = max(max_width, section_max_width)
    return max_width


import re
import datapie as dp

from rephorm.utility.report.range_utility import get_highlight, get_span


def get_table_highlight(table):

    start, end = get_highlight(table)
    span = get_span(table)
    skip_cols = 1
    if table.settings.show_units:
        skip_cols += 1

    length = len(span.start >> end)

    cols = range(list(span).index(start) + skip_cols, length + skip_cols)
    cols = list(cols)

    return cols

def prepare_row(
        series: dp.Series = None,
        span: dp.Span = None,
        title: str = "",
        unit: str = "",
        show_units: bool = None,
        decimal_precision: int = None,
        compare_style: str = None,
        footnote_references = None,
):
    if series is None:
        series_data = dp.Series()
        decimal_precision = 0

    else:
        series_data = series

    if isinstance(series_data, dp.Series):
        series_data = series_data[span].flatten().tolist()
    #if isinstance(series_data, TableSeries):
    else:
        series_data = series_data.data[span].flatten().tolist()


    if span is None:
        raise Exception("Span is missing.")

    options = {
        "diff": ("[", "]"),
        "pct": ("(", ")"),
    }

    left, right = options.get(
        compare_style,
        ("", ""),
    )

    # descriptions = [f"{spacer}{title}{footnote_references.get(i, ' ')}" for i in range(len(series_data))]

    descriptions = [f"{title}"]

    if show_units:
        descriptions = descriptions + [unit]


    pattern = r"\bnan\b|\[nan\]|\(nan\)"

    row = descriptions + [re.sub(pattern, "", f"{left}{x:.{decimal_precision}f}{right}") for x in series_data]

    return row


# make util funct cleanup later on, so less code regarding cleanup stuff. but call it here.
import os
import shutil

from rephorm.object_mappers._globals import reset_figure_map


def perform_cleanup(cleanup: bool = False, directory_path: str = None, base_file_path: str = None):

    reset_figure_map()
    # Reset background manager state between runs

    try:
        # Remove tmp dir where pdf figures are stored
        if cleanup and directory_path:
            shutil.rmtree(directory_path)
        # Remove base (initial) pdf
        if base_file_path and os.path.exists(base_file_path):
            os.remove(base_file_path)

    except Exception as e:
        print(f"Cleanup utility - error occurred: {e}")


from rephorm.objects.report import Report
from rephorm.objects.table_section import TableSection
from rephorm.objects.table import Table
from rephorm.objects.page_break import PageBreak
from rephorm.objects.grid import Grid
from rephorm.objects.chart import Chart
from rephorm.objects.chapter import Chapter
from .objects.text import Text
from rephorm.objects.footnote import Footnote
from rephorm.objects.chart_series import ChartSeries
from rephorm.objects.table_series import TableSeries
from .initialize import initialize_kaleido

__version__ = "1.1.16"
__author__ = 'OGResearch team'
__credits__ = 'OGResearch'
__license__ = 'Copyright 2025 OGResearch, all rights reserved'

initialize_kaleido()


from rephorm.object_mappers._utilities.grid_mapper_utility import compute_grid_layout
from rephorm.utility.report.add_footnotes import add_footnotes_at_bottom
from rephorm.utility.report.footnotes_counter import generate_footnote_numbers
from rephorm.utility.report.layout_utility import LayoutManager
from rephorm.utility.report.title_utility import add_title_with_references


class GridMapper:
    def __init__(self, grid):
        self.grid = grid

    def add_to_pdf(self, **kwargs):

        pdf = kwargs["pdf"]
        layout = self.grid.settings.layout
        nrow = self.grid.settings.nrow
        ncol = self.grid.settings.ncol
        footer_last_y = kwargs["footer_last_y"] if "footer_last_y" in kwargs else None
        starting_page = pdf.page_no()
        rs = self.grid.settings.styles["report"]
        footer_top_padding = rs["footer"]["top_padding"]
        footer_top_margin = rs["footer"]["top_margin"]
        title_bottom_padding = rs["title_bottom_padding"]
        # Todo: For grid why we dont take as in chapter? The used top height calculation?
        #  What is the reason to have Y coming from kwargs? Because then it takes Y from chapter/report layout calculations,
        #  and here it passes to layout... So I guess this is what offsets it differently, when adding chart to
        #  report/chapter directly and when having it inside 1x1 grid
        y = kwargs["y"]
        grid_size = self.grid.settings.ncol * self.grid.settings.nrow

        if layout is None:
            layout = compute_grid_layout(nrow=nrow, ncol=ncol)

        if len(self.grid.CHILDREN) > grid_size:
            raise Exception(
                f"Grid capacity exceeded: attempting to add {len(self.grid.CHILDREN)} object(-s) to a grid with capacity for only "
                f"{grid_size} object(-s) ({self.grid.settings.ncol}x{self.grid.settings.nrow}). "
                f"Please either increase grid size or reduce the number of objects."
            )

        footnote_references = generate_footnote_numbers(self.grid.footnotes)
        y = add_title_with_references(pdf, self.grid.title, footnote_references,
                                  self.grid.settings.styles["grid"]["title"]["font_family"],
                                  self.grid.settings.styles["grid"]["title"]["font_size"],
                                  self.grid.settings.styles["grid"]["title"]["font_style"], title_y=y)

        if bool(self.grid.footnotes):

                footer_last_y = add_footnotes_at_bottom(
                    pdf,
                    self.grid.footnotes,
                    footnote_references,
                    last_y=footer_last_y,
                    font_family=self.grid.settings.styles["footnotes"]["font_family"],
                    font_size=self.grid.settings.styles["footnotes"]["font_size"],
                    font_style=self.grid.settings.styles["footnotes"]["font_style"],
                    footer_padding=footer_top_padding,
                    footer_top_margin=footer_top_margin,
                )

        layout_manager = LayoutManager(
            pdf=pdf,
            used_top_height=y+title_bottom_padding,
        )

        positions = layout_manager.calculate_position(layout=layout, nrow=nrow, ncol=ncol)
        valid_positions = list(positions.keys())
        # Todo: set to the last chart | Do not remove, yet
        # last_element = valid_positions[-1]
        # last_element_data = positions[last_element]
        # pdf.line(pdf.l_margin, last_element_data["y"] + last_element_data["height"], pdf.w - pdf.r_margin, last_element_data["y"] + last_element_data["height"])

        for i, item in enumerate(self.grid.CHILDREN):

            position = valid_positions[i]
            pos_data = positions[position]

            current_page = pdf.page_no()
            if current_page > starting_page:
                starting_page = current_page
                footer_last_y = None


            grid_size = self.grid.settings.ncol * self.grid.settings.nrow

            mapper = item._get_mapper()
            mapper.add_to_pdf(
                pdf=pdf,
                x=pos_data["x"],
                y=pos_data["y"],
                w=pos_data["width"],
                h=pos_data["height"],
                grid_size=grid_size,
                footer_last_y=footer_last_y,
                context="grid",
            )

            # For debugging
            # pdf.set_draw_color(255, 0, 0)
            # pdf.rect(
            #     x=pos_data["x"],
            #     y=pos_data["y"],
            #     w=pos_data["width"],
            #     h=pos_data["height"],
            # )




from rephorm.dict.legend_positions import legend_positions
from rephorm.utility.report.resolve_color import resolve_color

def update_layout(title, settings, figure, img_w = None, img_h= None, y_range = None):

    font_size = settings.styles["chart"]["legend"]["font_size"]

    legend_pos = settings.legend_position.upper()

    # Normalize "rephorm" margin values to None so they fall through to defaults.
    margins = {
        k: (None if isinstance(v, str) and v.lower() == "rephorm" else v)
        for k, v in settings.styles["chart"]["margin"].items()
    }

    # Bottom margin auto-calculation:
    # Only when legend is at the bottom (SO, SEO, SWO) and user has not explicitly set a bottom margin for the legend.
    if legend_pos in ("SO", "SEO", "SWO") and margins["b"] is None:
        total = sum((len(t.name or "")+10)* font_size * 0.6 for t in figure.data)
        # img_w is essentially usable width for the chart. (In px)
        rows = -(-total // img_w)

        margins["b"] = (rows + 1) * (font_size + 8) + 20
        if img_h:
            # never exceed 30% of image height
            margins["b"] = min(margins["b"], img_h * 0.3)

    # Default to standard values if style margins are undefined (None).
    b_margin = margins["b"] or 0
    t_margin = margins["t"] or 5
    r_margin = margins["r"] or 0
    l_margin = margins["l"] or 0


    # Dict comprehension to include only margins that are not set to "plotly" in settings,
    # allowing Plotly to handle those automatically.
    final_margins = {
        k: v for k, v in {
            "b": b_margin,
            "t": t_margin,
            "r": r_margin,
            "l": l_margin
        }.items()
        if str(settings.styles["chart"]["margin"][k]).lower() != "plotly"
    }

    # Todo: extract this, but basically this is how to get title style in plotly for chart or for label too (I guess same way for label)
    font_style = settings.styles["chart"]["title"]["font_style"]
    if font_style == "B":
        formatted_title = f"<b>{title}</b>"
    elif font_style == "I":
        formatted_title = f"<i>{title}</i>"
    elif font_style == "BI":
        formatted_title = f"<b><i>{title}</i></b>"
    else:
        formatted_title = title

    layout = {
                "showlegend": settings.show_legend if hasattr(settings, "show_legend") else True,
                "title": {
                    "text": formatted_title,
                    "yref": "container",  # stop overlaps with plot area
                    "yanchor": "top", # align to top
                    "x": 0.5,
                    "automargin": True,  # adjust margins
                    "pad": {"b": 5},  # Fine control
                    "font_size": settings.styles["chart"]["title"]["font_size"],
                    "font_family": settings.styles["chart"]["title"]["font_family"],
                    "font_color": resolve_color(settings.styles["chart"]["title"]["font_color"]),
                },
                "plot_bgcolor": resolve_color(settings.styles["chart"]["bg_color"]),
                # paper_bgcolor is kinda useless for us now,
                # because of PDF image choice it will not support transparent bg
                # https://github.com/plotly/Kaleido/issues/91
                "paper_bgcolor": 'hsla(0, 100%, 50%, 0)',
                "margin": final_margins,
                "xaxis": {
                    # TODO: Later on introduce this parameter, to set number of ticks manually.
                    # "nticks": settings.styles["chart"]["x_axis"]["ticks"]["nticks"],
                    "showgrid": settings.show_grid if settings.show_grid else True,
                    "gridcolor": resolve_color(settings.styles["chart"]["grid_color"]),
                    "mirror": settings.axis_border,
                    "linecolor": resolve_color(settings.styles["chart"]["chart_border_color"]),
                    "showline": True,
                    "ticks": "inside",
                    "tickcolor": "black",
                    "linewidth": 1,
                    "tickwidth": 1,
                    "griddash": "dot",
                    "gridwidth": 1,
                    "tickfont": {"size": settings.styles["chart"]["x_axis"]["ticks"]["font_size"],
                                 "family": settings.styles["chart"]["x_axis"]["ticks"]["font_family"],
                                 "color": settings.styles["chart"]["y_axis"]["ticks"]["font_color"]},
                    "title": {
                        "text": settings.xaxis_title,
                        "font": {
                            "size": settings.styles["chart"]["x_axis"]["label"]["font_size"],
                            "family": settings.styles["chart"]["x_axis"]["label"]["font_family"]}},
                },
                "yaxis": {
                    "zeroline": settings.zeroline if hasattr(settings, 'zeroline') else True,
                    "showgrid": settings.show_grid if hasattr(settings, 'show_grid') else True,
                    "gridcolor": resolve_color(settings.styles["chart"]["grid_color"]),
                    "mirror": settings.axis_border,
                    "linecolor": resolve_color(settings.styles["chart"]["chart_border_color"]),
                    "showline": True,
                    "ticks": "inside",
                    "tickcolor": "black",
                    "linewidth": 1,
                    "tickwidth": 1,
                    "griddash": "dot",
                    "gridwidth": 1,
                    "range": y_range,
                    "tickfont": {"size": settings.styles["chart"]["y_axis"]["ticks"]["font_size"],
                                 "family": settings.styles["chart"]["y_axis"]["ticks"]["font_family"],
                                 "color": settings.styles["chart"]["y_axis"]["ticks"]["font_color"]},
                    "title": {
                        "text": settings.yaxis_title,
                        "font": {
                            "size": settings.styles["chart"]["y_axis"]["label"]["font_size"],
                            "family": settings.styles["chart"]["y_axis"]["label"]["font_family"]}},
                },
                "yaxis2": {
                    "overlaying": "y",
                    "side": "right",
                    "linecolor": resolve_color(settings.styles["chart"]["chart_border_color"]),
                    "showline": True,
                    "ticks": "inside",
                    "tickcolor": "black",
                    "linewidth": 1,
                    "tickwidth": 1,
                    "range": y_range,
                    "tickfont": {"size": settings.styles["chart"]["y_axis2"]["ticks"]["font_size"],
                                 "family": settings.styles["chart"]["y_axis2"]["ticks"]["font_family"],
                                 "color": settings.styles["chart"]["y_axis2"]["ticks"]["font_color"]},
                    "title": {
                        "text": settings.yaxis2_title,
                        "font": {
                            "size": settings.styles["chart"]["y_axis2"]["label"]["font_size"],
                            "family": settings.styles["chart"]["y_axis2"]["label"]["font_family"]}},
                },
                "legend": {
                    "font": {
                        "size": settings.styles["chart"]["legend"]["font_size"],
                        "family": settings.styles["chart"]["legend"]["font_family"],
                        "color": resolve_color(settings.styles["chart"]["legend"]["font_color"])
                    },
                    "orientation": settings.legend_orientation.lower(),
                    "x": legend_positions.get(settings.legend_position.upper()).get("x"),
                    "y": legend_positions.get(settings.legend_position.upper()).get("y"),
                    "xanchor": legend_positions.get(settings.legend_position.upper()).get("xanchor"),
                    "yanchor": legend_positions.get(settings.legend_position.upper()).get("yanchor"),
                    "borderwidth": settings.styles["chart"]["legend"]["border_width"],
                    "yref": legend_positions.get(settings.legend_position.upper()).get("yref"),
                    "xref": legend_positions.get(settings.legend_position.upper()).get("xref"),
                    "bgcolor": resolve_color(settings.styles["chart"]["legend"]["bg_color"]),

                }
            }

    # Update conditionally only if the user explicitly specifies 'ncols'.
    legend_ncol = settings.legend_ncol
    if bool(legend_ncol):
        figure.layout.legend["entrywidth"] = 1 / legend_ncol
        figure.layout.legend["entrywidthmode"] = "fraction"

    figure.update_layout(layout)


from logging import warning

from fpdf import FPDF

from rephorm.object_mappers._globals import get_figure_map
from rephorm.utility.report.cleanup_utility import perform_cleanup
from rephorm.utility.report.layout_utility import LayoutManager
from rephorm.utility.merge.pdf_merger import overlay_pdfs
from rephorm.utility.report.report_utility import set_title_page
from rephorm.object_mappers._globals import get_chart_queue, set_output_dir
import plotly.io as pio
import concurrent.futures
import os

# Platform-specific environment variables to stabilize Kaleido (>= 1.*) on Windows.
# These resolve the common "Wait expired / Browser watchdog" errors by:
# 1. KALEIDO_USE_SWIFT: Using SwiftShader for consistent software rendering.
# 2. KALEIDO_DISABLE_GPU: Preventing hangs/timeouts caused by missing or incompatible drivers.
os.environ["KALEIDO_USE_SWIFT"] = "true"
os.environ["KALEIDO_DISABLE_GPU"] = "true"

def _process_images_batch(task):
    """
    Worker process for Kaleido exports.
    task: tuple of (dim_key, batch_data) -> ("800x600", [...])
    """
    dim_key, batch_data = task
    if not batch_data:
        return 0

    figs = [item["fig"] for item in batch_data]
    paths = [item["path"] for item in batch_data]

    # Parse dimensions from the key (e.g., "800.0x600.0")
    try:
        w_str, h_str = dim_key.split('x')
        w = int(float(w_str)) if w_str != "None" else 800
        h = int(float(h_str)) if h_str != "None" else 600
    except (ValueError, AttributeError, TypeError):
        w, h = 800, 600

    # 'write_images' is more efficient even for lists of size 1
    pio.write_images(figs, paths, format="pdf", width=w, height=h)
    return len(figs)

class ReportMapper:

    def __init__(self, pdf: FPDF):
        self.pdf = pdf

    def compile(self, report, file_name: str, cleanup:bool = False):

        set_output_dir(os.path.dirname(os.path.abspath(file_name)) or ".")

        rs = report.settings.styles["report"]

        self.pdf.t_margin = rs["page_margin_top"]

        # TODO: Currently, we offset all bottom spacing to the bottom margin because FPDF2
        #  automatically breaks tables as soon as it reaches the bottom margin. This prevents
        #  any overlap with the footer, since the footer is included in fpdf.b_margin.
        #  We need to apply the same approach for the top margin and top spacing
        #  (e.g., header, title padding, etc.).
        # Also, make sure to verify whether any top padding or margin we have is actually
        # padding or margin. Padding refers to inner spacing, while margin is outer spacing.
        # Follow the standard convention.

        self.pdf.b_margin = rs["page_margin_bottom"] + rs["footer"]["height"] + rs["footer"]["top_margin"]
        self.pdf.l_margin = rs["page_margin_left"]
        self.pdf.r_margin = rs["page_margin_right"]
        self.pdf.c_margin = 0

        self.pdf.add_page()
        base_file_path = f"./tmp/report_tmp.pdf"
        directory_path = "./tmp/pdf_figures"

        used_top_height = rs["header"]["height"] + rs["header"]["bottom_margin"]

        layout_manager = LayoutManager(pdf=self.pdf, used_top_height=used_top_height)

        set_title_page(self.pdf, report)

        n_children = len(report.CHILDREN)

        for i, item in enumerate(report.CHILDREN):

            x, y, w, h = layout_manager.get_default_layout()
            mapper = item._get_mapper()
            # reset width to default
            # if children does not have width set by user, use the default width, if they do:
            # check if it does not exceed pdf.w and then use children's width instead.
            if hasattr(item, "width") and item.width is not None:
                if item.width > self.pdf.w:
                    warning(
                        f"{repr(item)} width: ({item.width}) | exceeds the PDF page width: ({self.pdf.w}). Falling back to defaults"
                    )
                else:
                    w = item.width

            mapper.add_to_pdf( pdf = self.pdf, w = w, h = h, x = x, y = y)

            if i < n_children-1:
                self.pdf.add_page()

        self.pdf.output(base_file_path)
        # Execute batch creation of PDFs synchronously

        # Phase 1: Parallel Image processing
        queue = get_chart_queue()
        all_tasks = []
        MAX_CHUNK_SIZE = 100

        for dim_key, charts in queue.items():
            for i in range(0, len(charts), MAX_CHUNK_SIZE):
                chunk = charts[i : i + MAX_CHUNK_SIZE]
                all_tasks.append((dim_key, chunk))

        if all_tasks:
            num_workers = min(len(all_tasks), 4)
            print(f"Parallelizing {len(all_tasks)} chunks across {num_workers} processes...")

            with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
                results = list(executor.map(_process_images_batch, all_tasks))

            print(f"Export complete. Total charts: {sum(results)}")

        # Phase 2: Merge images onto the PDF and cleanup
        overlay_pdfs(
            base_pdf_path=base_file_path,
            pdf_data=get_figure_map(),
            output_pdf_path=f"{file_name}.pdf"
        )

        perform_cleanup(cleanup, directory_path, base_file_path)




legend_positions = {
    # Center positions
    "C": {  # Center (Inline)
        "x": 0.5,
        "y": 0.5,
        "xanchor": "center",
        "yanchor": "middle",
        "yref": "paper"
    },
    # Inside positions
    "N": {  # North (Top-Center)
        "x": 0.5,
        "y": 1,
        "xanchor": "center",
        "yanchor": "top",
        "yref": "paper"
    },
    "S": {  # South (Bottom-Center)
        "x": 0.5,
        "y": 0,
        "xanchor": "center",
        "yanchor": "bottom",
        "yref": "paper"
    },
    "E": {  # East (Right-Center)
        "x": 1,
        "y": 0.5,
        "xanchor": "right",
        "yanchor": "middle",
        "yref": "paper"
    },
    "W": {  # West (Left-Center)
        "x": 0,
        "y": 0.5,
        "xanchor": "left",
        "yanchor": "middle",
        "yref": "paper"
    },
    "NE": {  # North-East (Top-Right)
        "x": 1,
        "y": 1,
        "xanchor": "right",
        "yanchor": "top",
        "yref": "paper"
    },
    "NW": {  # North-West (Top-Left)
        "x": 0,
        "y": 1,
        "xanchor": "left",
        "yanchor": "top",
        "yref": "paper"
    },
    "SE": {  # South-East (Bottom-Right)
        "x": 1,
        "y": 0,
        "xanchor": "right",
        "yanchor": "bottom",
        "yref": "paper"
    },
    "SW": {  # South-West (Bottom-Left)
        "x": 0,
        "y": 0,
        "xanchor": "left",
        "yanchor": "bottom",
        "yref": "paper"
    },

    # Outside positions
    "NO": {  # North-Outside (Above Top-Center)
        "x": 0.5,
        "y": 0.89,
        "xanchor": "center",
        "yanchor": "top",
        "yref": "container"
    },
    "SO": {
        "x": 0.5,
        "y": -0.1,
        "xanchor": "center",
        "yanchor": "top",
        "yref": "paper",
        "xref": "paper"
    },
    # South-Outside (ADJUSTED) Need proper testing
    # TODO: Consider changing back to "paper" and off-setting POS using negatives.
    # "SO": {
    #     "x": 0.5,
    #     "y": -0.15,
    #     "xanchor": "center",
    #     "yanchor": "top",
    #     "yref": "paper"
    # },
    "EO": {
        "x": 1.02,
        "y": 0.5,
        "xanchor": "left",
        "yanchor": "middle",
        "yref": "paper",
        "xref": "paper"
    },
    "WO": {  # West-Outside (Left of Chart)
        "x": 0,
        "y": 0.5,
        "xanchor": "left",
        "yanchor": "middle",
        "yref": "container",
        "xref": "container"
    },
    "NEO": {  # North-East-Outside (Above Top-Right)
        "x": 1,
        "y": -0.1,
        "xanchor": "right",
        "yanchor": "top",
        "yref": "container"
    },
    "NWO": {  # North-West-Outside (Above Top-Left)
        "x": 0,
        "y": -0.1,
        "xanchor": "left",
        "yanchor": "top",
        "yref": "container"
    },
    "SEO": {
    "x": 1,
    "y": 0,
    "xanchor": "right",
    "yanchor": "bottom",
    "yref": "container",
    },
    "SWO": {  # South-West-Outside (Below Bottom-Left)
        "x": 0,
        "y": 0,
        "xanchor": "left",
        "yanchor": "bottom",
        "yref": "container"
    }
}



from rephorm.utility.report.add_footnotes import add_footnotes_at_bottom
from rephorm.utility.report.footnotes_counter import generate_footnote_numbers
from rephorm.utility.report.layout_utility import LayoutManager
from rephorm.utility.report.title_utility import add_title_with_references


class ChapterMapper:
    def __init__(self, chapter):
        self.chapter = chapter

    # Todo: Reconsider add_to_pdf method naming. it would make more sense "render" or "map" because we do
    #  add things to the pdf and then go through the children here and eventually render them too...
    def add_to_pdf(self, **kwargs):

        pdf = kwargs["pdf"]
        footer_last_y = kwargs["footer_last_y"] if "footer_last_y" in kwargs else None
        starting_page = pdf.page_no()
        n_children = len(self.chapter.CHILDREN)


        rs = self.chapter.settings.styles["report"]
        used_top_height = rs["header"]["height"] + rs["header"]["bottom_margin"]
        footer_top_padding = rs["footer"]["top_padding"]
        footer_top_margin = rs["footer"]["top_margin"]

        pdf.set_y(used_top_height)

        pdf.set_font(
            family=self.chapter.settings.styles["chapter"]["font_family"],
            size=self.chapter.settings.styles["chapter"]["font_size"],
            style=self.chapter.settings.styles["chapter"]["font_style"],
        )

        if self.chapter.title is not None:
            footnote_references = generate_footnote_numbers(self.chapter.footnotes)
            title_y_pos = pdf.h / 2 - pdf.font_size / 2 # Todo: consider multiline title in the future (would need to calculate h of the text cell)
            add_title_with_references(pdf, self.chapter.title, footnote_references,
                                      font_family=self.chapter.settings.styles["chapter"]["title"]["font_family"],
                                      font_size=self.chapter.settings.styles["chapter"]["title"]["font_size"],
                                      font_style=self.chapter.settings.styles["chapter"]["title"]["font_style"],
                                      title_y=title_y_pos, cell_width=pdf.epw)
            if n_children > 0:
                pdf.add_page()

            if bool(self.chapter.footnotes):
                footer_last_y = add_footnotes_at_bottom(
                    pdf,
                    self.chapter.footnotes,
                    footnote_references,
                    last_y=footer_last_y,
                    font_family=self.chapter.settings.styles["footnotes"]["font_family"],
                    font_size=self.chapter.settings.styles["footnotes"]["font_size"],
                    font_style=self.chapter.settings.styles["footnotes"]["font_style"],
                    footer_padding=footer_top_padding,
                    footer_top_margin=footer_top_margin,
                )

        layout_manager = LayoutManager(
            pdf,
            used_top_height=used_top_height,
        )

        for i, item in enumerate(self.chapter.CHILDREN):

            x, y, width, height = layout_manager.get_default_layout()

            current_page_nr = pdf.page_no()
            if current_page_nr > starting_page:
                starting_page = current_page_nr
                # footer_last_y = None

            mapper = item._get_mapper()

            # TODO: IMPORTANT: Review the design and implement everything in a way where we pass single Y position
            #  across the object mappers, so we do not need Y - to set the position of chapter elements/objects,
            #  then title_y to correctly set the titles. We should pass Y along, and render elements/titles from the
            #  LAST known Y position (which is the last element rendered). This way we can avoid the need for
            #  title_y and footer_last_y, and god knows how many more Y's :D
            mapper.add_to_pdf(pdf=pdf, x=x, y=y, w=width, h=height, footer_last_y=footer_last_y)

            if i < n_children - 1:
                pdf.add_page()


import rephorm
import datapie as dp 
import numpy as np

# *****************************************
# 📌 DISCLAIMER
#
# Abbreviations in Code:
#    - `cs` → ChartSeries (for chart data)
#    - `ts` → TableSeries (for table data)
#    - Similar shorthand names may be used throughout the code.
#
# Random Data Notice:
#    - This script generates random values for demonstration.
#    - Replace `ran()` with actual data sources
#    in real use cases.
#
# NumPy (`np`) Usage:
#    - Functions prefixed with `np` come from the NumPy library.
#    - NumPy is only used to generate test data.
#
# Any code where comments starting with "⚠️ Generate" or "⚠️" pertains
# to generating sample test data. It is purely for simulation purposes
# and does not require interpretation.
#
# *****************************************

# ─────────────────────────────────────────
# ⚠️ Random Data Generator ⚠️
# It creates random values and is NOT needed in real applications.
ran = lambda length: np.cumsum(np.random.normal(0, 10, length))


# ─────────────────────────────────────────

def main():
    # ────── REPORT initialization ──────

    report = rephorm.Report(
        title="Advanced Reporting",
        orientation="L"
    )

    # ─────────────────────────────────────────
    # Table (Page 2)
    # ─────────────────────────────────────────

    # `show_units=False` - Disable display of units column in the table. (By default it is shown)
    # decimal_precision - number of decimal places to display for numerical values.
    # highlight - Defines a span (range) of values to be visually emphasized.
    # When `dp.end` is used in the `highlight` value, it automatically selects
    # the last date of the `span` assigned to the table.
    # The default highlight color is grey.

    # Customizing the default highlight color:
    # - Use the `styles` parameter, which accepts a dictionary of key-value pairs (refer to documentation for more details).
    # - The `table` key contains another dictionary since multiple style attributes can be defined for the table.
    # - Set the `"highlight_color"` key to `"green"` to override the default highlight color for the table.

    table_span = dp.qq(2022, 1) >> dp.qq(2025, 1)

    table = rephorm.Table(
        title="Random Data Table",
        show_units=False,
        span=table_span,
        decimal_precision=1,
        highlight=dp.qq(2024, 3) >> dp.end,
        styles={
            "highlight_color": "green"
        }
    )

    # Creating Time Series Data and Adding It to the Table

    # Generate a time series dataset (ts_data_1)
    # dp.Series (dp - DATAPIE) creates a time series with:
    #   dates=table_span: Time period covered by the table.
    #   values=ran(len(table_span)): Randomly generated numerical values representing sample data.
    ts_data_1 = dp.Series(dates=table_span, values=ran(len(table_span)))

    # Adding the time series data to the table using rephorm.TableSeries.
    # rephorm.TableSeries creates a rephorm-specific TableSeries object that takes dp.Series data.
    # The `title="series a"` parameter sets the row label for this data series in the table.
    # The `unit="units"` parameter defines the measurement unit displayed for this data series in the table.
    table.add(rephorm.TableSeries(
        title="series a",
        data=ts_data_1,
    ))

    ts_data_2 = dp.Series(dates=table_span, values=ran(len(table_span)))

    table.add(rephorm.TableSeries(
        title="series b",
        data=ts_data_2,
    ))

    ts_data_3 = dp.Series(dates=table_span, values=ran(len(table_span)))

    table.add(rephorm.TableSeries(
        title="series c",
        data=ts_data_3,
        
    ))

    ts_data_4 = dp.Series(dates=table_span, values=ran(len(table_span)))

    table.add(rephorm.TableSeries(
        title="series d",
        data=ts_data_4,
    ))

    ts_data_5 = dp.Series(dates=table_span, values=ran(len(table_span)))

    table.add(rephorm.TableSeries(
        title="series e",
        data=ts_data_5,
    ))

    ts_data_6 = dp.Series(dates=table_span, values=ran(len(table_span)))

    table.add(rephorm.TableSeries(
        title="series f",
        data=ts_data_6
    ))

    ts_data_7 = dp.Series(dates=table_span, values=ran(len(table_span)))

    table.add(rephorm.TableSeries(
        title="series g",
        data=ts_data_7,
    ))

    ts_data_8 = dp.Series(dates=table_span, values=ran(len(table_span)))

    table.add(rephorm.TableSeries(
        title="series h",
        data=ts_data_8,
    ))

    # ────── Add table section ──────

    # TABLE SECTION: what is it?
    # A table section is used to categorize and structure related data within
    # a table, enhancing organization and visual separation of data categories.
    # It introduces a special row (consisting of a table section title followed
    # by empty cells) to create a clear distinction between grouped data sets.
    # This improves readability and simplifies the interpretation of segmented
    # data in the table.

    # TABLE SECTION: Logic flow
    # The TableSeries should be added to the TableSection,
    # and then the TableSection is added to the table.

    # TABLE SECTION: Styles
    # Customizing the highlight color for a table section:
    # - Setting `"highlight_color"` to `"yellow"` changes the section's highlight color, and it
    # applies the same style to TableSeries rows that are later added to this section. (This is how inheritance works)

    table_section = rephorm.TableSection("Table Section i",
                                        styles={
                                            "highlight_color": "yellow"
                                        })

    # Add table series to table section
    table_section.add(rephorm.TableSeries(
        title="series j",
        data=ts_data_1,
    ))

    # TABLE SERIES: Comparison series
    # A comparison series behaves just like a normal TableSeries but with different styling.
    # - Visually, it appears directly below the last added row (or TableSeries).
    # - It is created by adding `comparison_series=True` to the TableSeries object
    # - Additionally, you can set `compare_style` to "pct" to enclose values in parentheses "()",
    #   or to "diff" to enclose values in square brackets "[]".

    table_section.add(rephorm.TableSeries(
        title="comp_series",
        data=dp.pct(ts_data_1, -4),
        compare_style="pct",
        comparison_series=True,
    ))

    table_section.add(rephorm.TableSeries(
        title="series k",
        data=ts_data_2,
    ))

    table_section.add(rephorm.TableSeries(
        title="comp_series",
        data=dp.pct(ts_data_2, -4),
        compare_style="diff",
        comparison_series=True,
    ))

    # Add table section to the table:
    table.add(table_section)

    # Add table to the report
    report.add(table)

    # ─────────────────────────────────────────
    #  CHARTS
    # ─────────────────────────────────────────

    # ─────────────────────────────────────────
    # Black line chart (Page 3)
    # ─────────────────────────────────────────

    # Define the Date Range for the Chart
    chart_span = dp.qq(2020, 1) >> dp.qq(2028, 1)

    # ⚠️ Generate Random Data sample
    blackline_chart_data = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    # Create Black line chart
    # `axis_border=True`: when it is set to `True`, it adds borders to the top and right sides,
    #  creating a fully enclosed frame around the chart.
    chart = rephorm.Chart(
        span=chart_span,
        title="Black line chart",
        highlight=dp.qq(2026, 1) >> dp.qq(2028, 1),
        show_legend=False,
        axis_border=True,
    )

    # Adding Time Series Data to the Chart A
    # - `rephorm.ChartSeries` creates a ChartSeries object that takes (dp.Series) time series data (`blackline_chart_data`).
    # - `series_type="line"` specifies that the data will be displayed as a line chart.
    # - The `styles` dictionary allows customization of the 'series' appearance.
    # - `"chart.series.line_color": "black"` sets the color of the line to black.
    # - In the code example below, the `styles` parameter is written in a flat key-value structure.

    chart.add(rephorm.ChartSeries(
        blackline_chart_data,
        series_type="line",
        styles={
            "line_color": "black",
        }
    ))

    # - Alternatively, the same configuration can be structured as a nested dictionary:
    #
    #   styles = {
    #       "chart": {
    #           "series": {
    #               "line_color": "black"
    #           }
    #       }
    #   }
    #
    # - Both formats achieve the same result, providing flexibility in defining styles.

    # Add Black line chart to the Report
    report.add(chart)

    # ─────────────────────────────────────────
    # Legend position (Page 4)
    # ─────────────────────────────────────────

    # Define the Date Range for the Chart
    chart_span = dp.qq(2015, 1) >> dp.qq(2035, 1)

    # ────── Configuring Legend Position & Orientation ──────

    # legend_position="EO",
    # - Defines where the legend appears relative to the chart.
    # - `"EO"` stands for "East Outside," placing the legend **outside** the chart on the right side.
    # - Other possible values include:
    #   - `"N"` (North, top inside the chart)
    #   - `"S"` (South, bottom inside the chart)
    #   - `"E"` (East, right inside the chart)
    #   - `"W"` (West, left inside the chart)
    #   - `"NO"` (North Outside), `"SO"` (South Outside), etc.

    # legend_orientation="V",
    # - Determines how the legend items are arranged.
    # - `"V"` (Vertical) stacks legend entries one below the other.
    # - `"H"` (Horizontal) places legend entries side by side.

    # This feature is still experimental.
    # - If `"EO"` (East Outside) or `"WO"` (West Outside) is selected for `legend_position`,
    #   it is necessary to set `legend_orientation="V"` (Vertical).
    # - Failing to do so may disrupt the layout of the chart, causing misalignment.

    chart = rephorm.Chart(
        span=chart_span,
        title="Legend position: EO, legend orientation: V",
        legend_position="EO",
        legend_orientation="V",
        show_grid=True,
    )

    # ⚠️ Generate a random sample data for the curve.
    series_1 = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    # Adding a time series to the chart:
    # - `rephorm.ChartSeries` creates a ChartSeries object using `series_1` time series data.
    # - `legend=("Custom",)` assigns a label for this series in the chart legend.
    # - `series_type="line"` specifies that the data should be plotted as a line chart.
    # - `markers_mode="lines"` ensures that only a continuous line is drawn, without markers.

    chart.add(rephorm.ChartSeries(
        series_1,
        legend=("Custom",),
        series_type="line",
        markers_mode="lines",
    ))

    # ⚠️ Generate a random sample data for the curve.
    series_2 = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    chart.add(rephorm.ChartSeries(
        series_2,
        series_type="line",
        markers_mode="lines",
        styles={
            "line_color": "black",
        }
    ))

    # Add the Chart to the Report
    report.add(chart)

    # ─────────────────────────────────────────
    # Different line styles (Page 5)
    # ─────────────────────────────────────────

    # Define the Date Range for the Chart
    chart_span = dp.qq(2015, 1) >> dp.qq(2035, 1)

    chart = rephorm.Chart(
        span=chart_span,
        title="Different line styles",
        legend_position="EO",
        legend_orientation="V",
        show_grid=True,
    )

    # Add Time Series Data with Different Line Styles

    # ⚠️ Generate a random sample data for the curve.
    series_1 = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    # Adding a time series to the chart:
    # - `rephorm.ChartSeries` creates a ChartSeries object using `series_1` time series data.
    # - `legend=("Custom",)` assigns a label for this series in the chart legend.
    # - `series_type="line"` specifies that the data should be plotted as a line chart.
    # - `markers_mode="lines"` ensures that only a continuous line is drawn, without markers.
    # - The `styles` dictionary customizes the appearance:
    #     - `"line_style": "dash"` makes the line dashed.
    #     - `"line_width": 2"` increase the thickness of the line. (default is 1)
    #     - `"line_color": "blue"` changes the line color to blue. (default is green)

    chart.add(rephorm.ChartSeries(
        series_1,
        legend=("dash",),
        series_type="line",
        markers_mode="lines",
        styles={
            "line_style": "dash",
            "line_width": 2,
            "line_color": "blue",
        }
    ))

    # ⚠️ Generate a random sample data for the curve.
    series_2 = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    chart.add(rephorm.ChartSeries(
        series_2,
        legend=("dot",),
        series_type="line",
        markers_mode="lines",
        styles={
            "line_style": "dot",
            "line_width": 2,
            "line_color": "orange",
        }
    ))

    # ⚠️ Generate a random sample data for the curve.
    series_3 = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    chart.add(rephorm.ChartSeries(
        series_3,
        legend=("dashdot",),
        series_type="line",
        markers_mode="lines",
        styles={
            "line_style": "dashdot",
            "line_width": 2,
            "line_color": "yellow",
        }
    ))

    # ⚠️ Generate a random sample data for the curve.
    series_4 = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    chart.add(rephorm.ChartSeries(
        series_4,
        legend=("longdash",),
        series_type="line",
        markers_mode="lines",
        styles={
            "line_style": "longdash",
            "line_width": 2,
            "line_color": "purple",
        }
    ))

    # ⚠️ Generate a random sample data for the curve.
    series_5 = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    chart.add(rephorm.ChartSeries(
        series_5,
        legend=("longdashdot",),
        series_type="line",
        markers_mode="lines",
        styles={
            "line_style": "longdashdot",
            "line_width": 2,
            "line_color": "magenta",
        }
    ))

    # Add the Chart to the Report
    report.add(chart)

    # ─────────────────────────────────────────
    # Single Marker Chart (Page 6)
    #
    # A chart where one of the time series (called "extension_series")
    # includes a marker, but only for the first data point.
    # ─────────────────────────────────────────

    # Specify Date Ranges
    main_span = dp.qq(2004, 1) >> dp.qq(2015, 2)
    extension_span = dp.qq(2010, 1) >> dp.qq(2015, 2)

    # Create Chart
    # - Instead of defining legends per series, `legend` can be set directly in `Chart`.
    chart = rephorm.Chart(
        title="Single marker Chart",
        span=main_span,
        axis_border=True,
    )

    # Creating Random Time Series Data Sample
    main_line_series = dp.Series(dates=main_span, values=ran(len(main_span)))

    # Adding Main Line Series
    chart.add(rephorm.ChartSeries(
        main_line_series,
        legend=("Main line",),
        series_type="line",
        markers_mode="lines",
        styles={
            "line_color": "blue",
        },
    ))

    # Creating Random Time Series Data Sample
    dashed_line_series = dp.Series(dates=extension_span, values=ran(len(extension_span)))

    # Adding Extension Line
    chart.add(rephorm.ChartSeries(
        dashed_line_series,
        legend=("Dashed line",),
        series_type="line",
        markers_mode="lines",
        styles={
            "line_color": "red",
            "line_style": "dash",
        },
    ))

    # Extract the first date & first value of extension series (for marker placement)
    first_date = dashed_line_series.dates[0]
    first_value = dashed_line_series.get_values()[0]

    # Create a single-point series for the marker
    first_point_series = dp.Series(dates=[first_date], values=[first_value])

    # Adds a single-point marker to the chart.
    # - show_legend=False: Hides the legend for this series.
    # - `series_type="line"`: Ensures compatibility within the chart.
    # - `markers_mode="markers"`: Displays only the markers.
    # - `chart.series.marker_size=10`: Sets the marker size.
    # - `chart.series.marker_width=2`: Sets the marker thickness.
    chart.add(rephorm.ChartSeries(
        first_point_series,
        show_legend=False,
        markers_mode="markers",
        styles={
            "marker_size": 10,
            "marker_width": 2,
            "marker_color": "red",
        },
    ))

    # Adding chart to the report
    report.add(chart)

    # ─────────────────────────────────────────
    # Contribution bar chart (Page 7)
    # ─────────────────────────────────────────

    # Define chart span
    plot_span = dp.qq(2014, 2) >> dp.qq(2022, 2)

    # Create chart object
    # The `<br>` tag within title string allows for a line break, enabling multi-line titles.
    # More than one <br> is not recommended. (Layout starts to break apart)
    shock_chart = rephorm.Chart(
        span=plot_span,
        title="Contribution bar <br> with multiline title",
        axis_border=True,
    )

    # Create and add bar series
    stacked_bar_series_1 = dp.Series(dates=plot_span, values=ran(len(plot_span)))
    shock_chart.add(rephorm.ChartSeries(
        stacked_bar_series_1,
        series_type="conbar",
        styles={
            "bar_face_color": "forestgreen",
        }
    ))

    stacked_bar_series_2 = dp.Series(dates=plot_span, values=ran(len(plot_span)))
    shock_chart.add(rephorm.ChartSeries(
        stacked_bar_series_2,
        series_type="conbar",
        styles={
            "bar_face_color": "darkgreen",
        }
    ))

    stacked_bar_series_3 = dp.Series(dates=plot_span, values=ran(len(plot_span)))
    shock_chart.add(rephorm.ChartSeries(
        stacked_bar_series_3,
        series_type="conbar",
        styles={
            "bar_face_color": "cadetblue",
        }
    ))

    stacked_bar_series_4 = dp.Series(dates=plot_span, values=ran(len(plot_span)))
    shock_chart.add(rephorm.ChartSeries(
        stacked_bar_series_4,
        series_type="conbar",
        styles={
            "bar_face_color": "olivedrab",
        }
    ))

    # ────── Compute the Element-Wise Sum ──────
    sum_series_values = np.add.reduce([
        stacked_bar_series_1.get_values(),
        stacked_bar_series_2.get_values(),
        stacked_bar_series_3.get_values(),
        stacked_bar_series_4.get_values()
    ])

    # Create Summed Line Series to Overlay on the Stacked Bars
    stacked_bar_line_series = dp.Series(dates=plot_span, values=sum_series_values)

    # ────── Adding a Border for the Line Series ──────
    # - rephorm does not support borders for line series natively. (Plotly neither)
    # - To simulate a border effect, we add two identical line series:
    #   1. The first line is plotted thicker in a neutral color to act as a "border."
    #   2. The second line, with a smaller width, is plotted on top to represent the actual data.
    #
    # The setup and data for both series are nearly identical.
    # The only differences are:
    #   1. `show_legend=False` for the first series, so it does not appear in the legend. (Important step)
    #   2. The `styles` parameters:
    #      - The first series has a thicker, neutral-colored line to create a "border" effect.
    #      - The second series represents the "actual" data, plotted on top with a distinct color and a thinner line.

    # Adding a thicker line in white color

    shock_chart.add(rephorm.ChartSeries(
        stacked_bar_line_series,
        series_type="line",
        show_legend=False,
        markers_mode="lines",
        styles={
            "line_width": 5,
            "line_color": "white",
        }
    ))

    # Adding the actual (thinner) line
    shock_chart.add(rephorm.ChartSeries(
        stacked_bar_line_series,
        series_type="line",
        markers_mode="lines",
        styles={
            "line_width": 3,
            "line_color": "grey",
        }
    ))

    # add chart to the report
    report.add(shock_chart)

    # ─────────────────────────────────────────
    # Line & Markers Chart (Page 8)
    # ─────────────────────────────────────────

    # ⚠️ Generate Random Data
    some_random_data = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    # Create Line & Markers Chart
    # - highlight: same as in tables, if you specify highlight parameter
    # and pass range/span as a value, it will add highlight (grey by default)

    lines_and_markers_chart = rephorm.Chart(
        span=chart_span,
        title="Line & Markers Chart",
        highlight=dp.qq(2030, 1) >> dp.end,
        axis_border=True,
    )

    # Add Line Series
    # marker_symbol:
    #  - Specifies the marker shape
    # List of possible symbols is available in documentation
    lines_and_markers_chart.add(rephorm.ChartSeries(
        some_random_data,
        show_legend=True,
        legend=("RedLine",),
        series_type="line",
        markers_mode="lines",
        marker_symbol="square-open",
        styles={
            "line_color": "darkred",
        }
    ))

    # Add "suqare-open" (dark green) markers to the chart
    lines_and_markers_chart.add(rephorm.ChartSeries(
        some_random_data,
        show_legend=True,
        legend=("square-open",),
        series_type="line",
        markers_mode="markers",
        marker_symbol="square-open",
        styles={
            "marker_color": "darkgreen",
            "marker_size": 6,
        }
    ))

    # Add "x-thin" (green) markers to the chart
    lines_and_markers_chart.add(rephorm.ChartSeries(
        some_random_data,
        show_legend=True,
        legend=("x-thin",),
        series_type="line",
        markers_mode="markers",
        marker_symbol="x-thin",
        styles={
            "marker_color": "mediumseagreen",
            "marker_size": 6,
            "line_width": 1,
        }
    ))

    # Add chart to the report
    report.add(lines_and_markers_chart)

    # ─────────────────────────────────────────
    # Working with styles (Page 9)
    # ─────────────────────────────────────────

    # This section will show how can user use styles differently and use to his own advantage.

    # The `combination_chart_style` dictionary below works like a design template for our chart.
    # By storing all our visual settings (colors, sizes, etc.) in one place, we can easily
    # apply consistent styling to the entire chart and all its parts without having to
    # copy-paste the same style code over and over again.

    # Chart-level properties define the overall visual style, while nested "series" properties
    # cascade down to all data series (ChartSeries objects) contained within the chart.
    # This approach lets us override default styles while keeping our code concise/minimal.

    # By "centralizing" styling, you can quickly update specific chart's appearance
    # without searching through scattered code. This saves time and keeps
    # your design consistent, especially when you have multiple elements.

    combination_grid = rephorm.Grid()

    combination_chart_style = {
        # Chart-level properties control the overall chart appearance:
        "highlight_color": "rgba(200, 200, 200, 0.5)",
        "grid_color": "lightblue",
        # Series-level properties automatically cascade to all chart child (ChartSeries)
        # objects within the chart:
        "series": {
            "line_width": 2,
            "line_color": "green",
            "marker_color": "green",
            "marker_size": 6
        }
    }

    # Generate random data for demonstration
    random_data = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    # Create first chart and apply the previously defined style template as:
    # `styles=combination_chart_style`
    combination_chart = rephorm.Chart(
        span=chart_span,
        highlight=dp.qq(2028, 2) >> dp.qq(2031, 4),
        styles=combination_chart_style
    )

    # Add a Chart Series to the chart.
    # This series inherits its styles from the parent chart, which means it automatically
    # follows the settings defined in the `combination_chart_style` dictionary.
    combination_chart.add(rephorm.ChartSeries(
        random_data,
    ))

    # Add chart to grid
    combination_grid.add(combination_chart)

    # Create second chart
    # apply same `combination_chart_style` styles dictionary
    combination_chart_2 = rephorm.Chart(
        span=chart_span,
        highlight=dp.qq(2028, 2) >> dp.qq(2031, 4),
        styles=combination_chart_style
    )

    # Generate random data for demonstration
    random_data_2 = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    # Add a Chart Series to the chart.
    combination_chart_2.add(rephorm.ChartSeries(
        random_data_2,
    ))

    # Add chart to grid
    combination_grid.add(combination_chart_2)

    # Create chart
    combination_chart_3 = rephorm.Chart(
        span=chart_span,
        highlight=dp.qq(2028, 2) >> dp.qq(2031, 4),
        styles=combination_chart_style
    )

    # Generate random data for demonstration
    random_data_3 = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    # Add a Chart Series to the chart.
    combination_chart_3.add(rephorm.ChartSeries(
        random_data_3,
    ))

    # Add chart to grid
    combination_grid.add(combination_chart_3)

    # Requirements: The last chart should have the same styling as defined in the  
    # combination_chart_style dictionary, but the series' line_color must be "Black".
    combination_chart_4 = rephorm.Chart(
        span=chart_span,
        markers_mode="lines",
        highlight=dp.qq(2028, 2) >> dp.qq(2031, 4),
        styles=combination_chart_style
    )

    # Generate random data for demonstration
    random_data_4 = dp.Series(dates=chart_span, values=ran(len(chart_span)))

    # Add a Chart Series to the chart.
    combination_chart_4.add(rephorm.ChartSeries(
        random_data_4,
        styles={
            "line_color": "black"
        }
    ))

    # Add chart to grid
    combination_grid.add(combination_chart_4)

    report.add(combination_grid)

    # The idea of this page 9 section, was to show that when multiple charts needs 
    # the same custom styling (different from the default), a styles dictionary 
    # could be created once and reused for each chart (of that type). Without this, 
    # the styles would have to be repeated for every chart instance. This approach 
    # keeps the code cleaner and follows the DRY (Don't Repeat Yourself) principle.

    # ─────────────────────────────────────────
    # Generating Report as a PDF
    # ─────────────────────────────────────────
    # Compile and export the final report to a PDF file
    # by calling the `.output` method on the `report` object,
    # which contains all the added elements.
    report.output("advanced_report")


if __name__ == "__main__":
    main()



from fpdf import FontFace

from rephorm.utility.report.resolve_color import resolve_color
from rephorm.utility.report.table_utility import prepare_row


class TableSeriesMapper:
    def __init__(self, series):
        self.series = series

    def add_to_pdf(self, **kwargs):


        highlight = kwargs["highlight"]
        pdf_table = kwargs["pdf_table"]
        span = kwargs["span"]
        show_units = kwargs["show_units"]
        #fpdf = kwargs["fpdf"]#todo: add to kwargs
        # last_y = kwargs["footer_last_y"] # We get last Y from parent,
        # so to add footnote to the bottom of the page right after parent footnotes.

        # Get start page nr before adding row
        # start_page_nr = fpdf.page_no()

        # add row
        element_row = pdf_table.row()

        #get page number after adding row
        # current_page_nr = fpdf.page_no()

        # In case the first element is on the next page, we need to reset last_y to none,
        # so that the add_footnotes_at_bottom function recalculates footnote positions
        # if current_page_nr > start_page_nr:
        #     last_y = None

        #Gen references here, pass it down to prep row and later to add_footnotes_. ...
        # footnote_references = generate_footnote_numbers(self.series.footnotes)

        row = prepare_row(
            series=self.series,
            compare_style=self.series.settings.compare_style,
            span=span,
            show_units=show_units,
            unit=self.series.unit,
            title=self.series.title,
            decimal_precision=self.series.settings.decimal_precision,
            footnote_references=None, # Todo: footnote_references
        )

        # add_footnotes_at_bottom(footnotes=self.series.footnotes,
        #     last_y=last_y,)

        color = resolve_color(self.series.settings.styles["table"]["series"]["highlight_color"])

        # I think we use similar logic in couple places, look if it's the case and put in utility.
        for i, cell in enumerate(row):  # Process each cell in the row

            style = FontFace(
                family=self.series.settings.styles["table"]["series"]["font_family"],
                emphasis=self.series.settings.styles["table"]["series"]["font_style"],
                size_pt=self.series.settings.styles["table"]["series"]["font_size"],
                fill_color= color if i in highlight else "#FFFFFF00",
                color=resolve_color(self.series.settings.styles["table"]["series"]["font_color"]),
            )

            # ^^^^^^^^^^^^^^^^^^^^^^^^^^

            if self.series.settings.comparison_series:
                padding = (-self.series.settings.styles["table"]["series"]["font_size"]*0.6, 0, 5, 0)
                style.size_pt *= 0.8 # Sets size of comparison series
            else:
                padding = (0, 0, 0, 0)

            element_row.cell(
                cell,
                style=style,
                padding = padding,
            )



import fpdf

def add_title_with_references(pdf, title, references, font_family, font_size, font_style, title_y, cell_width=None):

    # default fall back
    cell_width = cell_width or pdf.epw

    # anchor_x - starting x coordinate
    anchor_x = pdf.l_margin if cell_width == pdf.epw else pdf.get_x()

    pdf.set_font(font_family, font_style, font_size)

    actual_title_width = pdf.get_string_width(title)

    center_x = anchor_x + (cell_width - actual_title_width) / 2

    pdf.set_xy(center_x, title_y)

    # should set this on report level (report mapper)
    pdf.c_margin = 0

    is_multiline = "\n" in title

    # avoid auto-wrapping. Only break line if user specifies.
    if is_multiline:
        pdf.multi_cell(actual_title_width, None, title, align="C")
    else:
        pdf.cell(actual_title_width, None, title, align="C", new_y=fpdf.YPos.NEXT)

    y_after_title=pdf.y

    # Todo: References will not work if it is multiline and has newlines, because it sets ref_y to title_y, when for multiline it should set to last line start Y
    #  ref_x will not work as well ...

    if references:
        pdf.set_font_size(font_size*0.5)
        spacing = 3  # Space between references

        ref_x = center_x + actual_title_width + spacing

        # Set references
        for ref in references:
            ref_text = str(f"[{ref}]")
            pdf.set_xy(ref_x, title_y)
            pdf.cell(pdf.get_string_width(ref_text), None, ref_text, 0, 0, "L")
            ref_x += pdf.get_string_width(ref_text) + spacing

    # return last y pos
    return y_after_title # Todo: Check why not pdf.get_y() and pdf.y? Also check why I cant set last Y here and then use it later. why it breaks. Why we need to return this Y at all


import os
import sys
from logging import warning


def get_fonts(load_fonts):

    default_fonts = get_default_fonts()

    if not load_fonts:
        return default_fonts

    font_map = {(font['font_path']): font for font in default_fonts}

    # Update or append additional fonts
    for font in load_fonts:
        key = (font['font_path'])
        font_map[key] = font

    return list(font_map.values())

def get_default_fonts():
    base_path = os.path.abspath(os.path.dirname(__file__))
    return [
        {
         "font_family": "Open Sans",
         "font_path": os.path.join(base_path, "OpenSans.ttf"),
         "font_style": ""},
        {
         "font_family": "Open Sans",
         "font_path": os.path.join(base_path, "OpenSans-Italic.ttf"),
         "font_style": "I"},
        {
         "font_family": "Open Sans",
         "font_path": os.path.join(base_path, "OpenSans-Bold.ttf"),
         "font_style": "B"},
        {
         "font_family": "Open Sans",
         "font_path": os.path.join(base_path, "OpenSans-BoldItalic.ttf"),
         "font_style": "BI"},
    ]

def validate_fonts(fonts):

    required_styles = {"", "I", "B"}

    grouped_fonts = {}
    for font in fonts:
        family = font["font_family"]
        style = font["font_style"].upper()
        if family not in grouped_fonts:
            grouped_fonts[family] = set()
        grouped_fonts[family].add(style)

    for font_family, available_styles in grouped_fonts.items():
        missing_styles = required_styles - available_styles
        if missing_styles:
            raise Exception(f"Font family '{font_family}' is missing styles: {', '.join(missing_styles)}")


def add_custom_fonts(pdf, fonts):

    if fonts is not None:
        try:
            validate_fonts(fonts)
        except Exception as e:
            warning(e)
            sys.exit(1)

            # Font loading
    for font_dict in fonts:
        pdf.add_font(family = font_dict.get("font_family"),
                     fname=font_dict.get('font_path'),
                     style=font_dict.get('font_style', ''))



import os
import kaleido
import logging

"""
This module handles one-time setup tasks that cannot be reliably
executed during the standard 'pip install' phase due to modern
build constraints (like PEP 517 wheels).
"""

def initialize_kaleido():
    """
    Checks if Kaleido has been initialized for this specific installation.

    We only need to run "kaleido.get_chrome_sync()" once on a fresh
    package install. Here we use a sentinel file (".kaleido_initialized")
    stored inside the package directory to track this state.
    """
    package_dir = os.path.dirname(__file__)
    sentinel_file = os.path.join(package_dir, ".kaleido_initialized")

    if not os.path.exists(sentinel_file):
        try:
            print("rephorm: First-run initialization (downloading Kaleido binaries)...")
            kaleido.get_chrome_sync()
            # Create the flag file so this never runs again
            with open(sentinel_file, "w") as f:
                f.write("initialized")
        except Exception as e:
            logging.warning(f"rephorm: Could not initialize Kaleido: {e}")



from rephorm.dict.colors import color_codes


def resolve_color(color):

    normalized_color = color.lower()

    # Check if the input is within color_codes dict and if so - return hex code
    if normalized_color in color_codes:
        return color_codes[normalized_color]

    # Hex validation
    if normalized_color.startswith('#') and len(normalized_color) == 7:
        try:
            # raise a ValueError if not a valid hex
            int(normalized_color[1:], 16)
            return normalized_color
        except ValueError:
            pass

    return color

# This function is used to resolve the highlight color for chart
# When it comes in natural language such as "green", it resolves
# based on our set of colors.
# In color_codes the value must be in hex, cannot be in RGB.
def resolve_highlight_color(color, alpha=1):
    if color.lower().startswith(("#", "hsl", "rgb", "hsla", "rgba")):
        return color
    hex = resolve_color(color)
    return hex_to_rgba(hex, alpha=alpha)

def hex_to_rgba(hex_color, alpha=1):

    hex_color = hex_color.lstrip('#')
    r = int(hex_color[0:2], 16)
    g = int(hex_color[2:4], 16)
    b = int(hex_color[4:6], 16)
    return f"rgba({r}, {g}, {b}, {alpha})"


from rephorm.utility.is_set import is_set
footnote_counter = 0

def generate_footnote_numbers(footnotes):
    """
    Generate list of numbers for the footnotes
    Uses a global counter.
    """
    global footnote_counter
    references = []
    if is_set(footnotes):
        for _ in footnotes:
            footnote_counter += 1
            references.append(str(footnote_counter))
    return references



from typing import Literal
from rephorm.decorators.settings_validation import validate_kwargs
from rephorm.object_mappers.text_mapper import TextMapper
from rephorm.objects._utilities.settings_container import SettingsContainer


class Text:
    # ----------------------------------
    # Class initiator
    # ----------------------------------
    @validate_kwargs
    def __init__(
        self,
        title: str = "",
        text: str = "",
        align: Literal["x", "c", "l", "j", "r"] = "j",
        markdown = False,
        **kwargs
    ):
        self.title = title
        self.TEXT = text
        self.align = align
        self.markdown = markdown
        self.settings = SettingsContainer(**kwargs)

    def _get_mapper(self):
        return TextMapper(self)

    def __repr__(self):
        return f"{type(self).__name__}"








from datapie import Series as DataPieSeries

from rephorm.object_mappers.chart_series_mapper import ChartSeriesMapper
from rephorm.objects._utilities.settings_container import SettingsContainer
from rephorm.decorators.settings_validation import validate_kwargs
from rephorm.utility.add_style_prefix import add_prefix_to_styles


class ChartSeries:

    @validate_kwargs
    def __init__( self, data: DataPieSeries = None, **kwargs):
        """
            :param data (DataPieSeries, Required): Data for the chart series.
            :key yaxis (str, Optional): Y-axis position for the chart series. Defaults to "left".
            :key span (dp.Span): Span of the chart series.
            :key show_legend (bool): to show/hide the legend for this data series.
            :key legend (tuple): Specifies the legend labels for the series. For a single series, use ("Label",).
            For multivariate series, provide multiple labels like ("Label 1", "Label 2", ...).
            :key highlight (dp.Span): Span of the highlight.
            :key series_type (str): Type of series to display - "line" for line charts, "bar" for bar charts,
            "bar_contribution" for contribution bar charts; use "line" with markers_mode="markers" for marker-based charts.
            :key markers_mode (str): Display style for data points - "lines+markers" (lines with symbols/dots),
            "lines" (lines only), or "markers" (symbols/dots only).
            :key marker_symbol (str): Symbol used for markers in the series. e.g., "asterisk".
            :key update_traces (Dict): Traces configuration for the chart series. (Plotly specific) todo: check if this is used
            :key styles (Dict): Styles dictionary for additional customization (for details refer to report object).
        """

        self.data = data
        self.settings = SettingsContainer(**kwargs)

        if hasattr(self.settings, "styles"):
            self.settings.styles = add_prefix_to_styles("chart.series", self.settings.styles)

    def _get_mapper(self):
        return ChartSeriesMapper(self)

    def __repr__(self):
        return f"{type(self).__name__}"


import copy

from rephorm.dict.styles import default_styles, DEFAULT_FONT_FAMILY
from rephorm.object_params.settings import object_params
from rephorm.utility.merge.chart_properties_manager import get_bar_color, get_line_color, \
    get_line_style, update_chart_properties, get_line_width
from rephorm.utility.merge.merge_styles import update_nested_structure, create_nested_structure


def merge_settings(obj, params):
    """
    obj - The current object (e.g., Report, Chapter, etc.)
    params - A dictionary of parameters to propagate | That is report. settings
    """
    #Todo: Known issue: if you set report.font_size on level of chart,
    # it will not recalculate font sizes for chart and it's children.
    # ...
    # This is because at that point of iteration there is local params dictionary,
    # and we update it with what comes from children. To update reference
    # font format when children gets report.font_size parameter (reference parameters)
    # we would need to reinitialize default dict with those values from the children,
    # but we cant do it because we would loose the state of the local params,
    # which already have some important settings preset for us.
    # ...

    local_params = copy.deepcopy(params)

    parent_styles = create_nested_structure(obj.__class__.__name__, local_params.get("styles", {}))
    parent_font_family = parent_styles.get("report", {}).get("font_family", DEFAULT_FONT_FAMILY)

    object_styles = create_nested_structure(obj.__class__.__name__, obj.settings.__dict__.get("styles", {}))
    obj_font_family = object_styles.get("font_family", parent_font_family)

    # We update the default styles with styles coming from the previous parent (local_params["styles"]),
    # preserving inherited values. This avoids overwriting/resetting them as before (line 40),
    # where default_styles was directly assigned to local_params["styles"].

    local_params["styles"] = update_nested_structure(
        default_styles(obj_font_family),
        create_nested_structure(obj.__class__.__name__,local_params.get("styles", {}))
    )

    for key in object_params.keys():
        # Check if the key exists in params; if not, use its default value
        parent_value = local_params.get(key, None)
        # Check if obj.settings already has this key set
        if hasattr(obj.settings, key) and getattr(obj.settings, key) is not None:
            current_obj_value = getattr(obj.settings, key)

            if key == "styles":
                current_obj_value = update_nested_structure(parent_value, create_nested_structure(obj.__class__.__name__, current_obj_value))

            setattr(obj.settings, key, current_obj_value)
            local_params[key] = current_obj_value

        elif parent_value is not None:
            ultimate_owners = object_params.get(key, {}).get("ultimates", [])
            if type(obj).__name__ in ultimate_owners:
                setattr(obj.settings, key, parent_value)

        # If no value is set yet, use the default from ultimate_setting_owner
        elif getattr(obj.settings, key, None) is None:
            ultimate_owners = object_params.get(key, {}).get("ultimates", [])
            if type(obj).__name__ in ultimate_owners:
                if key == "styles":
                    setattr(obj.settings, key, getattr(local_params, key))
                else:
                    setattr(obj.settings, key, object_params[key].get("default_value", None))

    if obj.__class__.__name__ == "Chart":
        structured_styles = create_nested_structure(obj.__class__.__name__, obj.settings.styles) \
            if hasattr(obj.settings, "styles") \
            else {}
        update_chart_properties(structured_styles.get("chart", {}).get("bar_color_order", None),
                                structured_styles.get("chart", {}).get("line_color_order", None),
                                structured_styles.get("chart", {}).get("line_styles_order", None),
                                structured_styles.get("chart", {}).get("line_width_order", None),
                                )

    if hasattr(obj, "CHILDREN") and isinstance(obj.CHILDREN, list):
        list_of_obj = [obj.CHILDREN[-1]] if obj.__class__.__name__ == "Report" and obj.CHILDREN else obj.CHILDREN
        for child in list_of_obj:
            if hasattr(child, "settings"):
                merge_settings(child, local_params)
                handle_object_properties(child, child.settings.styles)

"""
This function process final styles dictionary and assigns colors to chartSeries based on series type, using predefined color order.
If the color is not user-defined, it assigns the next color from the global color order. (Non user defined colors coming as "None" by default)

When we call this function we do not need to ensure that the necessary dict structure
exists or that keys are present, because who calls this function should already have it set.
This function should not be called on styles dict, that was not processed (missing keys/values)
or that is not properly structured (dict that is flat / or not following the correct order)
"""

# TODO: Implement multivariate series support with consistent color assignment
#  Current Issue:
#  When creating a single series object that contains multiple data components (multivariate),
#  the color assignment system only applies a single color to the entire series.
#  Implementation:
#  Here we would detect if the series that is coming are multivariate or single,
#  basically check if single, assign color, else create an array for the length of
#  data columns in the series, populate it with get_bar_color() until it's full,
#  and pass it to styles as "bar_face_color"

def handle_object_properties(obj, merged_styles):

    if obj.__class__.__name__ == "ChartSeries":

        is_multivariate = obj.data.num_variants > 1

        series_type = getattr(obj.settings, "series_type")
        series_styles = merged_styles["chart"]["series"]

        def assign_style(key, generator_func):
            if key not in series_styles or series_styles[key] is None:
                if is_multivariate:
                    series_styles[key] = [generator_func() for _ in range(obj.data.num_variants)]
                else:
                    series_styles[key] = generator_func()

        if series_type in (
            "bar", "contribution_bar", "barcon", "conbar", "bar_color_stack",
            "bar_group", "bar_overlay", "bar_relative"
        ):
            assign_style("bar_face_color", get_bar_color)
        else:
            assign_style("line_color", get_line_color)
            assign_style("line_style", get_line_style)
            # line width: typically same for all?
            assign_style("line_width", get_line_width)





from rephorm.object_mappers.page_break_mapper import PageBreakMapper


class PageBreak:

    def __init__(self, ):
        pass

    def _get_mapper(self):
        return PageBreakMapper(self)

    def __repr__(self):
        return f"{type(self).__name__}"


DEFAULT_FONT_FAMILY = "Helvetica" # This was agreed upon

def default_styles(font_family=DEFAULT_FONT_FAMILY):
    return {
        "table": {
            "highlight_color": "#E0E0E0",
            "title": {
                "font_style": "B",
                "font_family": font_family,
                "font_color": "#000000",
                "font_size": 13,
            },
            "heading": {
                "font_style": "",
                "font_family": font_family,
                "font_color": "#000000",
                "font_size": 10,
                "highlight_color": "#E0E0E0",
            },
            "series": {
                "font_style": "",
                "font_family": font_family,
                "font_color": "#000000",
                "font_size": 10,
                "highlight_color": "#E0E0E0",
            },
            "section": {
                "font_style": "",
                "font_family": font_family,
                "font_color": "#000000",
                "font_size": 10,
                "highlight_color": "#E0E0E0",
            }
        },
        "text": {
            "font_style": "",
            "font_family": font_family,
            "font_size": 10,
            "title": {
                "font_style": "B",
                "font_family": font_family,
                "font_size": 16,
            }
        },
        "footnotes": {
            "font_size": 8,
            "font_color": "#000000",
            "font_style": "",
            "font_family": font_family,
        },
        "grid": {
            "title": {
                "font_style": "B",
                "font_family": font_family,
                "font_color": "#000000",
                "font_size": 13,
            },
        },
        "chapter": {
            "title": {
                "font_style": "B",
                "font_family": font_family,
                "font_color": "#000000",
                "font_size": 16,
            },
            "font_style": "",
            "font_family": font_family,
            "font_size": 11,
        },
        "chart": {
            "margin": {
                "t": None,
                "r": None,
                "l": None,
                "b": None,
            },
            "highlight_color": "rgba(0, 0, 0, 0.12)",
            "grid_color": "lightgrey",
            "bg_color": "rgba(0, 255, 0, 0)",
            "chart_border_color": "#000000",
            "line_styles_order": [
                "solid"
            ],
            "line_color_order": [
                "rgb(31, 119, 180)",
                "rgb(255, 127, 14)",
                "rgb(44, 160, 44)",
                "rgb(214, 39, 40)",
                "rgb(148, 103, 189)",
                "rgb(140, 86, 75)",
                "rgb(227, 119, 194)",
                "rgb(127, 127, 127)",
                "rgb(188, 189, 34)",
                "rgb(23, 190, 207)"
            ],
            "bar_color_order": [
                "rgb(31, 119, 180)",
                "rgb(255, 127, 14)",
                "rgb(44, 160, 44)",
                "rgb(214, 39, 40)",
                "rgb(148, 103, 189)",
                "rgb(140, 86, 75)",
                "rgb(227, 119, 194)",
                "rgb(127, 127, 127)",
                "rgb(188, 189, 34)",
                "rgb(23, 190, 207)"
            ],
            "line_width_order": [1],
            "series": {
                "bar_edge_color": "black",
                "bar_face_color": None,
                "bar_edge_width": 1.0,
                "line_width": None,
                "line_style": None,
                "line_color": None,
                "marker_color": None,
                "marker_size": 1,
                "marker_width": 1,
            },
            "legend": {
                "border_width": 0,
                "bg_color": "rgba(255, 255, 255, 0)",
                "font_style": "",
                "font_family": font_family,
                "font_color": "#000000",
                "font_size": 8,
            },
            "title": {
                "font_style": "",
                "font_family": font_family,
                "font_color": "#000000",
                "font_size": 11,
            },
            "x_axis": {
                "ticks": {
                    "font_size": 10,
                    "font_color": "#000000",
                    "font_family": font_family,
                },
                "label": {
                    "font_size": 10,
                    "font_color": "#000000",
                    "font_family": font_family,
                }
            },
            "y_axis": {
                "ticks": {
                    "font_size": 10,
                    "font_color": "#000000",
                    "font_family": font_family,
                },
                "label": {
                    "font_size": 10,
                    "font_color": "#000000",
                    "font_family": font_family,
                },
            },
            "y_axis2": {
                "ticks": {
                    "font_size": 10,
                    "font_color": "#000000",
                    "font_family": font_family,

                },
                "label": {
                    "font_size": 10,
                    "font_color": "#000000",
                    "font_family": font_family,
                }
            },
        },
        "report": {
            # only front page title
            "title": {
                "font_style": "B",
                "font_size": 25,
                "font_color": "#000000",
                "font_family": font_family,
            },
            "subtitle": {
                "font_style": "B",
                "font_size": 21,
                "font_color": "#000000",
                "font_family": font_family,
            },
            "abstract": {
                "height": 50,
                "font_size": 10,
                "font_color": "#000000",
                "font_family": font_family,
            },
            "footer": {
                "height": 15,
                "top_margin": 5,
                "top_padding": 5,
            },
            "header": {
                "height": 20,
                "bottom_margin": 5,
            },
            "font_style": "",
            "font_family": font_family,
            "font_color": "#000000",
            "page_margin_top": 10,
            "page_margin_right": 40,
            "page_margin_bottom": 15,
            "page_margin_left": 40,
            #Todo: Find a better name for it.
            "title_bottom_padding": 10,
            "logo_height": 15,
        }
    }


import rephorm
import datapie as dp
import numpy as np


# *****************************************
# 📌 DISCLAIMER
#
# Abbreviations in Code:
#    - `cs` → ChartSeries (for chart data)
#    - `ts` → TableSeries (for table data)
#    - Similar shorthand names may be used throughout the code.
#
# Random Data Notice:
#    - This script generates random values for demonstration.
#    - Replace `ran()` with actual data sources in real use cases.
#
# NumPy (`np`) Usage:
#    - Functions prefixed with `np` come from the NumPy library.
#    - NumPy is only used to generate test data.
#
# *****************************************


def main():
    # *****************************************
    # ⚠️ DEMO CODE - Random Data Generator ⚠️
    # It creates random values and is NOT
    # needed in real-world applications.
    ran = lambda: np.random.uniform(10, 100, 9)
    # *****************************************

    # ─────────────────────────────────────────
    # REPORT INITIALIZATION
    # ─────────────────────────────────────────

    # Initialize the report object with a title, subtitle, and landscape orientation.
    report = rephorm.Report(
        title="Hello World",
        subtitle="Subtitle Space",
        orientation="L"
    )

    # Now, we define and add various elements to the report.

    # ─────────────────────────────────────────
    # TEXT (Page 2)
    # ─────────────────────────────────────────

    # Create a text object and adding some textual data
    text = rephorm.Text(
        "In economic theory, the concepts of supply and demand form the backbone of market equilibrium. "
        "The law of demand states that, all else being equal, as the price of a good increases, "
        "the quantity demanded by consumers decreases, and vice versa. This inverse relationship "
        "is driven by consumer purchasing power and the availability of substitutes."
    )

    # Add the text object to the report
    report.add(text)

    # ─────────────────────────────────────────
    # TABLE (Page 3)
    # ─────────────────────────────────────────

    # Define a table:
    # First, we specify the date range for the table.
    start_date = dp.qq(2023, 1)
    end_date = dp.qq(2025, 1)

    # Create a quarterly time span for the table:
    # The >> operator in DataPie generates a sequential time range between two
    # specified dates.
    table_span = start_date >> end_date

    # Initialize a table object with the defined time span
    # We explicitly assign `table_span` to the `span` parameter to ensure correct mapping.
    # Directly passing `table_span` without specifying `span` would cause errors
    # since the constructor expects multiple arguments.
    table = rephorm.Table(span=table_span)

    # After initializing the table object, we need to populate it with data.
    # The TableSeries object is used to store and structure time series data within the table.

    # Create a TableSeries object to hold data for the table:
    # Each TableSeries corresponds to a single row in the table.
    ts_1 = rephorm.TableSeries(
        data=dp.Series(dates=table_span, values=ran()),
        title="Series 1",
        unit="mil"
    )
    # Add table series to table.
    table.add(ts_1)

    # Add more "rows" (aka, TableSeries) to the table:
    ts_2 = rephorm.TableSeries(
        data=dp.Series(dates=table_span, values=ran()),
        title="Series 2",
        unit="mil"
    )
    table.add(ts_2)

    ts_3 = rephorm.TableSeries(
        data=dp.Series(dates=table_span, values=ran()),
        title="Series 3",
        unit="mil"
    )
    table.add(ts_3)

    ts_4 = rephorm.TableSeries(
        data=dp.Series(dates=table_span, values=ran()),
        title="Series 4",
        unit="mil"
    )
    table.add(ts_4)

    ts_5 = rephorm.TableSeries(
        data=dp.Series(dates=table_span, values=ran()),
        title="Series 5",
        unit="mil"
    )
    table.add(ts_5)

    # Add the table to the report.
    report.add(table)

    # ─────────────────────────────────────────
    # CHART (Page 4)
    # ─────────────────────────────────────────

    # First, we specify the date range for the chart.
    start_date = dp.qq(2023, 1)
    end_date = dp.qq(2025, 1)

    # Create a quarterly time span for the chart:
    # The >> operator in DataPie Toolbox generates a sequential time range between
    # two specified dates.
    chart_span = start_date >> end_date

    # Initialize the chart object with the specified time span.
    # Chart object can hold one or more data series (ChartSeries) for visualization.
    chart = rephorm.Chart(span=chart_span)

    # Create a ChartSeries object:
    # This represents one dataset (one plotted line) in the chart.
    # It takes a time series as input, with random values assigned to each date.
    cs_1 = rephorm.ChartSeries(data=dp.Series(dates=chart_span, values=ran()))

    # Add the created ChartSeries (data) to the chart.
    chart.add(cs_1)

    report.add(chart)

    # ─────────────────────────────────────────
    # GRID (Page 5)
    # ─────────────────────────────────────────

    # initialize Grid object
    # When no parameters added, it will default to 2x2 grid
    # Otherwise pass "ncol" - number of columns
    # And "nrow" - number of rows, to create custom grid layouts
    grid = rephorm.Grid(ncol=1, nrow=2)

    # Initializing chart
    grid_chart_1 = rephorm.Chart(span=chart_span)

    # Create a ChartSeries object:
    cs_2 = rephorm.ChartSeries(
        data=dp.Series(dates=chart_span, values=ran()),
        series_type="contribution_bar"
    )

    # Add the created ChartSeries (data) to the chart.
    grid_chart_1.add(cs_2)

    # Adding chart o the grid
    grid.add(grid_chart_1)

    # Initializing chart
    grid_chart_2 = rephorm.Chart(span=chart_span)

    # Create a ChartSeries object to represent a data series in the chart.
    #
    # This series will use **markers only** (without lines or bars),
    # making only the data points visible.
    #
    # The `styles` parameter is used to customize the appearance:
    # - `"chart.series.marker_size": 8` increases the marker size,
    #   because the default size would be too small and hard to see.
    #   (styles are covered more in advanced_report.py)
    cs_3 = rephorm.ChartSeries(
        data=dp.Series(dates=chart_span, values=ran()),
        markers_mode="markers",
        styles={"marker_size": 8}
    )

    # Add the created ChartSeries (data) to the chart.
    grid_chart_2.add(cs_3)

    # Adding chart o the grid
    grid.add(grid_chart_2)

    # Adding grid to the report
    report.add(grid)

    # ─────────────────────────────────────────
    # Generating Report as a PDF
    # ─────────────────────────────────────────
    # Compile and export the final report to a PDF file
    # by calling the `.output` method on the `report` object,
    # which contains all the added elements.
    report.output("basic_report")


if __name__ == "__main__":
    main()


# Matlab default colors: https://nl.mathworks.com/help/matlab/creating_plots/specify-plot-colors.html
# Keep codes in hex format here, because FontFace do not accept rgb or rgba, (and we use it for text)
# for charts and plotly we do conversion to rgba in resolve_highlight_color

color_codes = {
    "red": "#FF0000",
    "r": "#FF0000",
    "green": "#00FF00",
    "g": "#00FF00",
    "blue": "#0000FF",
    "b": "#0000FF",
    "cyan": "#00FFFF",
    "c": "#00FFFF",
    "magenta": "#FF00FF",
    "m": "#FF00FF",
    "yellow": "#FFFF00",
    "y": "#FFFF00",
    "black": "#000000",
    "k": "#000000",
    "white": "#FFFFFF",
    "w": "#FFFFFF"
}



from fpdf import FPDF, enums

# Todo: The right place for this is mapper package.
def get_page_dimensions(pdf: FPDF):
    return pdf.w, pdf.h

# Todo: The right place for this is mapper package.
# Todo: 03/31/2025: Kinda useless function. This could be done in the report mapper.

def set_title_page(pdf: FPDF, report_obj):

    bool(report_obj.title) and set_title(pdf, report_obj)
    bool(report_obj.abstract) and set_abstract(pdf, report_obj)
    pdf.add_page()

def set_title(pdf: FPDF, report_obj):

    title_height = report_obj.settings.styles["report"]["title"]["font_size"]

    space_between = title_height

    title_y_pos = pdf.eph / 2 - title_height / 2

    pdf.set_y(title_y_pos)

    pdf.set_font(report_obj.settings.styles["report"]["title"]["font_family"],
                 report_obj.settings.styles["report"]["title"]["font_style"],
                 report_obj.settings.styles["report"]["title"]["font_size"])

    pdf.cell(
        text=report_obj.title,
        w=0,
        align="C",
        # Moves cursor to new ln, same behavior as in multi_cell,
        # Necessary to get same space between, when using pdf.ln
        new_y=enums.YPos.NEXT
    )

    pdf.ln(space_between)

    pdf.set_font(report_obj.settings.styles["report"]["subtitle"]["font_family"],
                 report_obj.settings.styles["report"]["subtitle"]["font_style"],
                 report_obj.settings.styles["report"]["subtitle"]["font_size"])

    pdf.multi_cell(
        text=report_obj.subtitle,
        w=0,
        align="C",
    )
    pdf.ln(space_between)

    if bool(report_obj.settings.logo):
        logo_height = report_obj.settings.styles["report"]["logo_height"] # Todo: should get extracted to logo height - styles/params.!! Extract to styles
        pdf.image(
            report_obj.settings.logo,
            x=enums.Align.C,
            y=pdf.get_y(),
            h=logo_height,
        )

    # reset the font
    pdf.set_font(
        family = report_obj.settings.styles["report"]["font_family"],
        style = report_obj.settings.styles["report"]["font_style"],
    )

def set_abstract(pdf: FPDF, report_obj):

    rs = report_obj.settings.styles["report"]
    abstract_y_pos = pdf.h - pdf.b_margin - rs["abstract"]["height"]
    abstract_font_size= rs["abstract"]["font_size"]
    abstract_text = report_obj.abstract

    line_height = 1.5
    pdf.set_font(family=rs["abstract"]["font_family"], size=abstract_font_size)

    pdf.set_y(abstract_y_pos)

    pdf.set_draw_color(0, 0, 0)
    pdf.line(x1=pdf.l_margin, y1=abstract_y_pos - 5, x2=pdf.w-pdf.r_margin, y2=abstract_y_pos - 5)

    b_margin = pdf.b_margin

    pdf.set_auto_page_break(auto=False)

    pdf.multi_cell(
        text=abstract_text,
        w=0,
        h = abstract_font_size * line_height, # Todo consider extracting this 1.5 multiplier
        align="l", # Todo: Consider extracting to styles
    )

    pdf.set_auto_page_break(auto=True, margin=b_margin)



"""
pdf.b_margin = b.margin + footer_height + footer_top_margin.
"""
class LayoutManager:
    def __init__(self, pdf, used_top_height = 0):
        """
        Initialize the layout manager with pdf instance, and header and footer height.
        """
        self.pdf = pdf
        self.used_top_height = used_top_height

    def calculate_position(self, layout, nrow, ncol):
        """
        Here we compute grid layout based on user inputs and return array with every element X Y W H "coordinates"
        Result - array with coordinates of each element and key is (row_indx, col_indx)
        """
        result={}

        cell_width = ((self.pdf.w - self.pdf.l_margin - self.pdf.r_margin) / ncol)
        # b_margin = b_margin + footer_height + footer_top_margin
        cell_height = (self.pdf.h - self.pdf.b_margin - self.used_top_height) / nrow

        for row_indx in range (nrow):
            for col_indx in range(ncol):
                cell = layout[row_indx][col_indx]

                if cell is None:
                    continue

                rowspan = cell.get("rowspan", 1)
                colspan = cell.get("colspan", 1)

                width = colspan * cell_width
                height = rowspan * cell_height

                x = col_indx * cell_width + self.pdf.l_margin # sets from left
                y = row_indx * cell_height + self.used_top_height # sets from top

                result[(row_indx, col_indx)] = {
                    "x": x,
                    "y": y,
                    "width": width,
                    "height": height
                }

        return result

    def get_default_layout(self):
        """
        Gets the position (x, y) and available size (width, height) for a chapter or report,
        considering headers and footers to determine the exact placement of its content (child object).
        """
        available_width = self.pdf.w - self.pdf.l_margin - self.pdf.r_margin
        available_height = self.pdf.h - self.pdf.t_margin - self.pdf.b_margin - self.used_top_height

        x = self.pdf.l_margin
        y = self.used_top_height + self.pdf.t_margin

        return x, y, available_width, available_height



from datapie import Series as IrisSeries

from rephorm.object_mappers.table_series_mapper import TableSeriesMapper
from rephorm.objects._utilities.settings_container import SettingsContainer
from rephorm.decorators.settings_validation import validate_kwargs
from rephorm.utility.add_style_prefix import add_prefix_to_styles


class TableSeries:

    @validate_kwargs
    def __init__(self, title: str = None, unit: str = None, data: IrisSeries = None, **kwargs):
        """
            :param data: (IrisSeries, Required) Data for the table series.
            :param title: (str) Name of the table series / row.
            :param unit: (str) The text value for the "units" column in this specific row of the table series.
            This defines what will be displayed as the unit for the corresponding row.
            :key span (dp.Span): Span of the table series.
            :key highlight (dp.Span): Span of the highlight in the table series.
            :key compare_style (str): Comparison style for the table series. "diff", "pct".
            :key decimal_precision (int): Sets number of decimal places after the decimal point.
            :key comparison_series (bool): Whether the series should be displayed as comparison series.
            :key styles (Dict): Styles dictionary for additional customization (for details refer to report object).
        """
        self.data = data
        self.title = title
        self.unit = unit
        self.settings = SettingsContainer(**kwargs)
        if hasattr(self.settings, "styles"):
            self.settings.styles = add_prefix_to_styles("table.series", self.settings.styles)

    def _get_mapper(self):
        return TableSeriesMapper(self)

    def __repr__(self):
        return f"{type(self).__name__}"


from __future__ import annotations
import copy
from typing import Optional, Union, List

from rephorm.objects.table import Table
from rephorm.objects.text import Text
from rephorm.objects.chart import Chart
from rephorm.object_mappers.grid_mapper import GridMapper
from rephorm.objects._utilities.settings_container import SettingsContainer
from rephorm.decorators.settings_validation import validate_kwargs
from rephorm.utility.add_style_prefix import add_prefix_to_styles


class Grid:

    @validate_kwargs
    def __init__(self, title: str = "", footnotes: Optional[List[str]] = None, **kwargs):
        """
            :param title: (str) Title of the grid.
            :param footnotes: (Tuple(str)) Footnotes for the title.
            :key ncol (int): Number of columns in the grid.
            :key nrow (int): Number of rows in the grid.
            :key layout (Dict): layout configuration for the chart (see documentation).
            :key styles (Dict): Styles dict (refer to documentation).
        """
        self.title = title
        self.CHILDREN = []
        self.footnotes = footnotes
        self.settings = SettingsContainer(**kwargs)
        if hasattr(self.settings, "styles"):
            self.settings.styles = add_prefix_to_styles("grid", self.settings.styles)

    def add(
        self,
        grid_child : Union[Grid, Table, Text, Chart] = None,
    ):
        #12/10/2024: We should have checks in the objects and not object mappers
        # We would also not have problem with imports here, because this module itself is within OBJECTS
        # Todo: Update latest docs, explaining it is not possible to nest grids within grids, and the reason why.
        if grid_child is not None and isinstance(grid_child, (Table, Grid, Text, Chart)):
            # grid_child._on_add_callback()
            copy_grid_child = copy.deepcopy(grid_child)
            self.CHILDREN.append(copy_grid_child)
        else: raise Exception("Grid: grid child of wrong type or None")


    def _get_mapper(self):
        return GridMapper(self)

    def __repr__(self):
        return f"{type(self).__name__}"

    def _on_add_callback(self):
        pass
    # This would be called by whoever is adding the grid object.


from datetime import datetime

from fpdf import FPDF

"""
This subclass extends FPDF to include customized header and footer functionality,
making it easier to apply consistent formatting across all pages of the document.
By overriding the `header` and `footer` methods, we ensure that these elements
are automatically added to each page without manual intervention.
"""

class PDF(FPDF):
    def __init__(self, report_styles = None, report_title = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.report_styles = report_styles
        self.report_title = report_title
        self.timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    def header(self):
        header_height = self.report_styles["header"]["height"]
        self.set_font(family=self.report_styles["font_family"], size=8)
        # Todo: here, H should be set to header height
        # Set absolute position for the header
        # We need to skip first cell (bcs of logo).
        header_width = self.epw / 3
        y = self.get_y()

        layout = self._get_header_layout(self.page_no(), self.timestamp, self.report_title)

        for item in layout:
            x = self.l_margin + item["slot"] * header_width
            self.set_xy(x, y)
            self.cell(
                w=header_width,
                h=header_height,
                text=item["text"],
                align=item.get("align", "L"),
                # border=1 #for debugging
            )
    # Todo: was easier to extract it like this, and in the future we could rewr this function
    #  further in order to specify exact positions of elements.
    def _get_header_layout(self, page_no, timestamp, report_title):
        layout = [
            {
                "id": "title",
                "text": report_title,
                "align": "L",
                "slot": 0
            },
            {
                "id": "page",
                "text": f"Page {page_no}",
                "align": "C",
                "slot": 1
            },
            {
                "id": "timestamp",
                "text": timestamp,
                "align": "R",
                "slot": 2
            }
        ]

        # Todo: modify this, once we have show_title_page option on report object that controls whether the
        #  title page shown or not.
        if page_no == 1:
            layout = [{
                "id": "timestamp",
                "text": timestamp,
                "align": "R",
                "slot": 2
            }]

        return layout









import copy
from typing import Optional, Union, List

from rephorm.object_mappers.chapter_mapper import ChapterMapper
from rephorm.objects.chart import Chart
from rephorm.objects.page_break import PageBreak
from rephorm.objects.grid import Grid
from rephorm.objects._utilities.settings_container import SettingsContainer
from rephorm.objects.table import Table
from rephorm.objects.text import Text
from rephorm.decorators.settings_validation import validate_kwargs
from rephorm.utility.add_style_prefix import add_prefix_to_styles


class Chapter:
    @validate_kwargs
    def __init__(self, title: str = None, footnotes: Optional[List[str]] = None, **kwargs):
        """
            :param title: (str) Title of the chapter.
            :key styles (Dict): Styles dictionary for additional customization (for details refer to report object).
        """
        self.title = title
        self.footnotes = footnotes
        self.CHILDREN = []
        # if "styles" in kwargs:
        #     kwargs["styles"]=adjust_styles(kwargs["styles"], caller)
        self.settings = SettingsContainer(**kwargs)
        if hasattr(self.settings, "styles"):
            self.settings.styles = add_prefix_to_styles("chapter", self.settings.styles)

    def add(
        self,
        chapter_child: Union[Grid, Table, Text, PageBreak, Chart] = None,
    ):
        if chapter_child is not None and isinstance(chapter_child, (Grid, Table, Text, PageBreak, Chart)):
            copy_chapter_child = copy.deepcopy(chapter_child)
            self.CHILDREN.append(copy_chapter_child)
        else: raise Exception("Chapter: chapter child of wrong type or None")

    def _get_mapper(self):
        return ChapterMapper(self)

    def __repr__(self):
        return f"{type(self).__name__}"


from typing import Dict
import datapie as dp
from rephorm.dict.styles import default_styles

"""
    This Dictionary defines all possible parameters that can be applied to any object within the report.

    - default_value - default value of the parameter
    - type - for type checking | if type is None, then it is not being checked ("useful" in case of Complex data Types)
    - ultimates - list of objects that should get/receive this parameter
"""

object_params = {
    "span": {
        "default_value": dp.start >> dp.end,
        "type": dp.Span,
        "ultimates": ["Chart", "Table", "ChartSeries", "TableSeries"],
    },
    "highlight": {
        "default_value": None,
        "type": dp.Span,
        "ultimates": ["Chart", "Table", "TableSection", "TableSeries"],
    },
    "show_legend": {
        "default_value": True,
        "type": bool,
        "ultimates": ["ChartSeries", "Chart"],
    },
    "yaxis": {
        "default_value": "left",
        "type": str,
        "possible_values":
            {"left", "right"},
        "ultimates": ["ChartSeries"],
    },
    "update_traces": {  # todo: check if this is used
        "default_value": None,
        "type": Dict,
        "ultimates": ["ChartSeries"],
    },
    "show_grid": {
        "default_value": True,
        "type": bool,
        "ultimates": ["Chart"],
    },
    "axis_border": {
        "default_value": False,
        "type": bool,
        "ultimates": ["Chart"],
    },
    "xaxis_title": {
        "default_value": "",
        "type": str,
        "ultimates": ["Chart"],
    },
    "yaxis_title": {
        "default_value": "",
        "type": str,
        "ultimates": ["Chart"],
    },
    "yaxis2_title": {
        "default_value": "",
        "type": str,
        "ultimates": ["Chart"],
    },
    "legend_orientation": {
        "default_value": "h",
        "type": str,
        "ultimates": ["Chart"],
    },
    "legend_position": {
        "default_value": "SO",
        "type": str,
        "possible_values":
            {"N", "S", "E", "W",
             "NE", "NW", "SE", "SW",
             "NO", "SO", "EO", "WO",
             "NEO", "NWO", "SEO", "SWO"},
        "ultimates": ["Chart"],
    },
    "ncol": {
        "default_value": 2,
        "type": int,
        "ultimates": ["Grid"],
    },
    "nrow": {
        "default_value": 2,
        "type": int,
        "ultimates": ["Grid"],
    },
    "apply_report_layout": {
        "default_value": False,
        "type": bool,
        "ultimates": ["Chart"],
    },
    "layout": {
        "default_value": None,
        "type": None,
        "ultimates": ["Grid"],
    },
    "comparison_series": {
        "default_value": False,
        "type": bool,
        "ultimates": ["TableSeries"],
    },
    "series_type": {
        "default_value": "line",
        "type": str,
        "possible_values": {"line", "bar", "contribution_bar", "barcon", "conbar", "bar_stack", "bar_group",
                            "bar_overlay", "bar_relative"},
        "ultimates": ["Chart", "ChartSeries"],
    },
    "markers_mode": {
        "default_value": "lines",
        "type": str,
        "possible_values": {"lines+markers", "lines", "markers"},
        "ultimates": ["ChartSeries"],  # TEST IT: Chart was removed as Chart should not use it in any way
    },
    "legend": {
        "default_value": None,
        "type": tuple,
        "ultimates": ["Chart", "ChartSeries"],
    },
    "compare_style": {
        "default_value": "",
        "type": str,
        "ultimates": ["TableSeries"],
    },
    "marker_symbol": {
        "default_value": "asterisk",
        "type": str,
        "ultimates": ["Chart", "ChartSeries"],
    },
    "orientation": {
        "default_value": "P",
        "type": str,
        "ultimates": ["Report"],
    },
    "unit": {
        "default_value": "pt",
        "type": str,
        "ultimates": ["Report"],
    },
    "format": {
        "default_value": "A4",
        "type": str,
        "ultimates": ["Report"],
    },
    "decimal_precision": {
        "default_value": 1,
        "type": int,
        "ultimates": ["TableSeries"],
    },
    "show_units": {
        "default_value": True,
        "type": bool,
        "ultimates": ["TableSection", "Table"],
    },
    "zeroline": {
        "default_value": False,
        "type": bool,
        "ultimates": ["Chart"],
    },
    "legend_ncol": {
        "default_value": None,
        "type": int,
        "ultimates": ["Chart"],
    },
    "frequency": {
        "default_value": None,
        "type": None,  # Cant check this one
        "ultimates": ["Table"],
    },
    "styles": {
        "default_value": default_styles(),
        "type": None,  # TODO: change to Dict later on
        "ultimates": ["Report", "Chapter", "TableSection", "ChartSeries", "TableSeries", "Table", "Chart", "Grid",
                      "Text"],
    },
    "logo": {
        "default_value": "",
        "type": str,
        "ultimates": ["Report"],
    },
    "save_as": {
        "default_value": "",
        "type": str,
        "ultimates": ["Chart"],
    }
}


from rephorm.dict.update_traces import update_base_traces, apply_per_trace_styles


class ChartSeriesMapper:
    def __init__(self, series):
        self.series = series

    def add_to_pdf(self, **kwargs):

        figure = kwargs["figure"]
        num_variants = self.series.data.num_variants
        # Here we will get either single value or a list of values, depends if multivariate or not:
        line_colors = self.series.settings.styles["chart"]["series"]["line_color"]
        line_widths = self.series.settings.styles["chart"]["series"]["line_width"]
        line_styles = self.series.settings.styles["chart"]["series"]["line_style"]
        bar_colors = self.series.settings.styles["chart"]["series"]["bar_face_color"]

        def get_series_type(key):
            mapping = {
                "line": "line",
                "bar": "bar",
                "bar_stack": "bar_stack",
                "contribution_bar": "bar_relative",
                "conbar": "bar_relative",
                "barcon": "bar_relative",
                "bar_relative": "bar_relative",
                "bar_group": "bar_group",
                "bar_overlay": "bar_overlay",
            }
            return mapping.get(key)

        fig = self.series.data.plot(
            span=self.series.settings.span,
            date_format_style="compact",
            date_axis_mode="instant",
            figure=figure,
            chart_type=get_series_type(self.series.settings.series_type),
            legend=self.series.settings.legend,
            show_legend=self.series.settings.show_legend,
            include_in_legend=self.series.settings.show_legend,
            return_info=True,
            show_figure=False,
            update_traces=update_base_traces(self.series.settings)
        )

        apply_per_trace_styles(
            fig["figure"],
            line_colors,
            line_widths,
            line_styles,
            bar_colors,
            num_variants
        )


def compute_grid_layout(nrow = 1, ncol = 1):
    """
    Computes the layout of a grid based on the number of rows and columns.
    """
    layout = []
    for indx in range(nrow):
        row_layout = []
        for indx in range(ncol):
            row_layout.append({})
        layout.append(row_layout)
    return layout


import os

import datapie
import numpy as np
from plotly.graph_objects import Figure

from rephorm.object_mappers._globals import set_figure_map
from rephorm.object_mappers._utilities.chart_mapper_utility import process_bar_data
from rephorm.dict.update_layout import update_layout
from rephorm.utility.report.resolve_color import resolve_highlight_color
from rephorm.object_mappers._globals import add_to_chart_queue, get_output_dir



class ChartMapper:

    def __init__(self, chart):
        self.chart = chart
        self.figure = Figure()
        self.apply_report_layout = chart.settings.apply_report_layout

    def add_to_pdf(self, **kwargs):
        grid_size = kwargs.get("grid_size", 1)
        pdf = kwargs["pdf"]
        x = kwargs["x"]
        y = kwargs["y"]
        w = kwargs["w"]
        h = kwargs["h"]

        scale_f = pdf.k

        img_w = w * scale_f
        img_h = h * scale_f

        if grid_size > 9:
            img_w = None
            img_h = None

        # Todo: Directory creation and check, should be extracted from here, too many concerns...
        directory = "./tmp/pdf_figures"
        os.makedirs(directory, exist_ok=True)

        # set figure for rendering, either external (which we would get from self.chart.figure or internal rephorm)
        # or internal
        self.figure = self.chart.figure or self._add_internal_figure(w, h)

        # Apply rephorm report layout to external figure if apply_report_layout is true
        if self.apply_report_layout and bool(self.chart.figure):
            update_layout(self.chart.figure.layout.title.text, self.chart.settings, self.figure, w, h)

        #Picture exporting | Saving to specific path if save_as is provided,
        # otherwise saving to temp directory with unique name based on chart id. (That gets cleaned up)
        # Todo: This should be extracted from here, too many concerns in this function.
        save_as = self.chart.settings.save_as
        if save_as:
            if not save_as.lower().endswith(".pdf"):
                save_as += ".pdf"
            file_path = save_as if os.path.isabs(save_as) else os.path.join(get_output_dir(), save_as)
            os.makedirs(os.path.dirname(os.path.abspath(file_path)), exist_ok=True)
        else:
            file_path = f"{directory}/{id(self.chart)}.pdf"

        add_to_chart_queue({
            "fig": self.figure.to_dict(),
            "path": file_path,
            "width": img_w,
            "height": img_h
        })

        fig_map = ({"page": pdf.page_no(),
                    "pdf_path": file_path,
                    "x": x * scale_f,
                    "y": y * scale_f,
                    "width": w * scale_f,
                    "height": h * scale_f})

        set_figure_map(fig_map)

    def _add_internal_figure(self, img_w, img_h):

        min_y = None
        max_y = None

        span = self.chart._get_span()
        span_length = len(span)
        neg_stacked_bars_data = np.zeros(span_length)
        pos_stacked_bars_data = np.zeros(span_length)
        stacked_bars = False

        for item in self.chart.CHILDREN:
            for variant in range(1, item.data.num_variants + 1):
                y_values = item.data.get_data_variant_from_until(span, variant).flatten()
                if item.settings.series_type in ("barcon", "conbar", "contribution_bar", "bar_relative"):
                    stacked_bars = True
                    neg_stacked_bars_data, pos_stacked_bars_data = process_bar_data(
                        neg_stacked_bars_data,
                        pos_stacked_bars_data,
                        y_values
                    )
                else:
                    # Direct min/max calculation for non-barcon series
                    current_min = y_values.min()
                    current_max = y_values.max()

                    min_y = current_min if min_y is None else min(min_y, current_min)
                    max_y = current_max if max_y is None else max(max_y, current_max)

            series_mapper = item._get_mapper()
            series_mapper.add_to_pdf(
                figure=self.figure)  # Todo: This add_to_pdf call is confusing, yes, we wanted to unify function names, but in this case it does not make sense.

        if stacked_bars:
            min_y = min(neg_stacked_bars_data) if min_y is None else min(min_y, min(neg_stacked_bars_data))
            max_y = max(pos_stacked_bars_data) if max_y is None else max(max_y, max(pos_stacked_bars_data))

        dy = max_y - min_y
        y_range = [min_y - dy * 0.01, max_y + dy * 0.01]

        update_layout(self.chart.title, self.chart.settings, self.figure, img_w, img_h, y_range)

        # Adjusts view SPAN on charts. "Slices" any padding between.
        # self.figure.layout.xaxis.range = [span.start.to_python_date(position="start"),
        #                                   span.end.to_python_date(position="end")]

        if self.chart.settings.highlight is not None and datapie is not None:
            datapie.ez_plotly.highlight(
                self.figure,
                self.chart._get_highlight(),
                color=resolve_highlight_color(
                    self.chart.settings.styles["chart"]["highlight_color"], alpha=0.25
                ),
            )

            # Todo: This is a workaround for the issue with the highlight. It should be fixed in the irispie library.
            # therefore, delete update_shapes() call here, when it's fixed in iris-pie
            self.figure.update_shapes({"layer": "below"})

        return self.figure



def mm_to_px(mm, ppi):
    """Converts millimeters to pixels given the pixels per inch (PPI)."""
    return (mm / 25.4) * ppi
    # return mm * 3.78

# Todo: make generic function that converts report units (whatever we get) to respective ones we need / to px


#pt is what overlay_pdfs uses
def mm_to_pt(mm):
    """Converts millimeters to points."""
    return mm * 2.83465


import copy

from rephorm.dict.styles import default_styles

class ChartPropertiesManager:
    def __init__(self):
        self.init_properties()

    def init_properties(self):

        # actually we don't need default here bcs it takes from styles in merge_settings anyway.
        self.default_bar_colors = default_styles()["chart"]["bar_color_order"]
        self.default_line_colors = default_styles()["chart"]["line_color_order"]
        self.default_line_styles = default_styles()["chart"]["line_styles_order"]
        self.default_line_widths = default_styles()["chart"]["line_width_order"]
        # set memory
        self.bar_colors = copy.deepcopy(self.default_bar_colors)
        self.line_widths = copy.deepcopy(self.default_line_widths)
        self.line_colors = copy.deepcopy(self.default_line_colors)
        self.line_styles = copy.deepcopy(self.default_line_styles)

        # set working stack
        self.bar_color_stack = copy.deepcopy(self.bar_colors)
        self.line_widths_stack = copy.deepcopy(self.line_widths)
        self.line_color_stack = copy.deepcopy(self.line_colors)
        self.line_style_stack = copy.deepcopy(self.line_styles)

    def get_bar_color(self):
        if not self.bar_color_stack:
            self.bar_color_stack = copy.deepcopy(self.bar_colors)
        return self.bar_color_stack.pop(0)

    def get_line_width(self):
        if not self.line_widths_stack:
            self.line_widths_stack = copy.deepcopy(self.line_widths)
        return self.line_widths_stack.pop(0)

    def get_line_color(self):
        if not self.line_color_stack:
            self.line_color_stack = copy.deepcopy(self.line_colors)
        return self.line_color_stack.pop(0)

    def get_line_style(self):
        if not self.line_style_stack:
            self.line_style_stack = copy.deepcopy(self.line_styles)
        return self.line_style_stack.pop(0)

    def update_colors_for_chart(self, bar_colors=None, line_colors=None, line_styles=None, line_widths=None):

        self.bar_color_stack.clear()
        self.line_color_stack.clear()
        self.line_style_stack.clear()
        self.line_widths_stack.clear()

        self.bar_colors = bar_colors if bar_colors and bar_colors is not None else copy.deepcopy(self.default_bar_colors)
        self.line_colors = line_colors if line_colors and line_colors is not None else copy.deepcopy(self.default_line_colors)
        self.line_styles = line_styles if line_styles and line_styles is not None else copy.deepcopy(self.default_line_styles)
        self.line_widths = line_widths if line_widths and line_widths is not None else copy.deepcopy(self.default_line_widths)

        self.bar_color_stack = copy.deepcopy(self.bar_colors)
        self.line_color_stack = copy.deepcopy(self.line_colors)
        self.line_style_stack = copy.deepcopy(self.line_styles)

chart_properties = ChartPropertiesManager()

def get_bar_color():
    return chart_properties.get_bar_color()

def get_line_width():
    return chart_properties.get_line_width()

def get_line_color():
    return chart_properties.get_line_color()


def get_line_style():
    return chart_properties.get_line_style()

def update_chart_properties(bar_colors=None, line_colors=None, line_styles=None, line_widths=None):
    return chart_properties.update_colors_for_chart(bar_colors, line_colors, line_styles, line_widths)


import copy

from rephorm.dict.styles import default_styles

# Get the allowed Styles structure from default_styles
allowed_structure = default_styles()

def validate_value_type(obj_name, path, value, structure=allowed_structure):
    keys = path.split('.')
    final_structure = structure
    current_path = []

    for key in keys:
        current_path.append(key)
        if isinstance(final_structure, dict):
            if key not in final_structure:
                raise KeyError(f"({obj_name} object): Styles key '{key}' for '{'.'.join(current_path[:-1])}' does not exist. "
                               f"Available keys at this level: {list(final_structure.keys())}")
            final_structure = final_structure.get(key, {})
        else:
            # If we've reached a non-dictionary node, we can't go further; this means wrong nesting
            raise KeyError(f"Styles: Invalid path '{path}'. No further nesting allowed under '{keys[-2]}'.")

    if isinstance(final_structure, dict):
        if not isinstance(value, dict):
            available_keys = list(final_structure.keys())
            raise ValueError(
                f"Styles key error: '{path}' expects a dictionary with keys: {available_keys}. Got {type(value).__name__} instead.")
        else:
            # Recursively validate each sub-key for dictionaries
            for sub_key, sub_value in value.items():
                new_path = f"{path}.{sub_key}"
                validate_value_type(obj_name, new_path, sub_value)
    else:
        # Skips type validation if the type of Final structure's (edge parameter in default_styles)
        # value was set to None.
        if type(final_structure) is type(None):
            return

        expected_type = type(final_structure) if not isinstance(final_structure, type) else final_structure

        if not isinstance(value, expected_type) and not (isinstance(value, (int, float)) and expected_type in (int, float)):
            raise TypeError(
                f"Styles: Invalid type for '{path}': Expected {expected_type.__name__}, got {type(value).__name__}")

def create_nested_structure(obj_name, flat_dict):

    if flat_dict is None:
        return flat_dict

    nested_dict = {}

    flat_dict = copy.deepcopy(flat_dict)

    for key, value in flat_dict.items():
        validate_value_type(obj_name, key, value, allowed_structure)
        keys = key.split(".")
        d = nested_dict

        for k in keys[:-1]:
            d = d.setdefault(k, {})

        d[keys[-1]] = value

    return nested_dict

def update_nested_structure(org_dict, new_dict):
    if new_dict is None:
        return org_dict

    result = copy.deepcopy(org_dict)

    for key, value in new_dict.items():
        if isinstance(value, dict) and key in result and isinstance(result[key], dict):
            result[key] = update_nested_structure(result[key], value)
        else:
            result[key] = value

    return result


# Central storage for globals
figure_map = []
chart_queue = {}
_output_dir = "."

def set_figure_map(data):
    global figure_map
    figure_map.append(data)

def get_figure_map():
    global figure_map
    return figure_map

def add_to_chart_queue(chart_data):
    """Groups charts by width x height for batch processing."""
    global chart_queue
    w, h = chart_data.get("width"), chart_data.get("height")

    # Ensure dimensions are strings for the key
    key = f"{w}x{h}"

    if key not in chart_queue:
        chart_queue[key] = []

    chart_queue[key].append({
        "fig": chart_data["fig"],
        "path": chart_data["path"]
    })

def get_chart_queue():
    global chart_queue
    return chart_queue

def reset_figure_map():
    global figure_map, chart_queue
    figure_map = []
    chart_queue = {}

def set_output_dir(path):
    global _output_dir
    _output_dir = path

def get_output_dir():
    return _output_dir


"""
Generic settings container.
Encapsulates all settings for a specific object.
"""
class SettingsContainer:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


from rephorm.utility.report.title_utility import add_title_with_references


class TextMapper:
    def __init__(self, text):
        self.text = text

    def add_to_pdf(self, **kwargs):
        pdf = kwargs["pdf"]
        w = kwargs["w"]
        y = kwargs["y"]
        x = kwargs["x"]

        family = self.text.settings.styles["text"]["title"]["font_family"]
        size = self.text.settings.styles["text"]["title"]["font_size"]
        style = self.text.settings.styles["text"]["title"]["font_style"]

        # todo: check multipliers everywhere and use line height const for it.
        line_height = 1.5 # todo: consider extracting to styles, once again.

        y = add_title_with_references(pdf=pdf, title=self.text.title, references=None,
                                  font_family=family, font_size=size, font_style=style, title_y=y)

        pdf.set_font(family=self.text.settings.styles["text"]["font_family"],
                     size=self.text.settings.styles["text"]["font_size"],
                     style=self.text.settings.styles["text"]["font_style"])

        pdf.set_xy(x, y)

        pdf.multi_cell(
            w=w,
            h=self.text.settings.styles["text"]["font_size"] * line_height,
            txt=self.text.TEXT,
            align=self.text.align,
            markdown=self.text.markdown,
        )



from fpdf import FontFace
from rephorm.utility.report.resolve_color import resolve_color

class TableSectionMapper:
    def __init__(self, table_section):
        self.table_section = table_section

    def add_to_pdf(self, **kwargs):

        n_cols = kwargs["n_cols"]
        highlight = kwargs["highlight"]
        pdf_table = kwargs["pdf_table"]
        span = kwargs["span"]
        title_row = pdf_table.row()


        for i in range(n_cols):

            color = resolve_color(self.table_section.settings.styles["table"]["section"]["highlight_color"])

            style = FontFace(
                emphasis= self.table_section.settings.styles["table"]["section"]["font_style"],
                size_pt=self.table_section.settings.styles["table"]["section"]["font_size"],
                fill_color= color if i in highlight else "#FFFFFF00",
                family=self.table_section.settings.styles["table"]["section"]["font_family"],
                color=self.table_section.settings.styles["table"]["section"]["font_color"],
            )

            content = self.table_section.TITLE if i == 0 else ""
            title_row.cell(content, style=style)

        for item in self.table_section.CHILDREN:
            mapper = item._get_mapper()
            mapper.add_to_pdf(span=span, pdf_table=pdf_table, highlight=highlight, show_units=self.table_section.settings.show_units)










import copy
from typing import Optional, Dict, Text, Union

from rephorm.objects.grid import Grid
from rephorm.objects.table import Table
from rephorm.objects.text import Text
from rephorm.objects.page_break import PageBreak
from rephorm.objects.chart import Chart
from rephorm.objects.chapter import Chapter

from rephorm.objects._utilities.settings_container import SettingsContainer
from rephorm.utility.add_style_prefix import add_prefix_to_styles
from rephorm.utility.fonts.font_loading import get_fonts, add_custom_fonts
from rephorm.object_mappers.report_mapper import ReportMapper
from rephorm.utility.PDF import PDF
from rephorm.utility.merge.merge_settings import merge_settings
from rephorm.decorators.settings_validation import validate_kwargs


class Report:
    @validate_kwargs
    # Todo: delete report name from here, it is not needed since you can specify it on output method call
    def __init__(self, title: str = "", subtitle: str = "", load_fonts: Optional[list[Dict[str, str]]] = None, abstract: str="", **kwargs):
        # Todo: consider moving `load_fonts` to `kwargs` (Currently, it would be hard to validate entry/value if we move to kwargs, bcs of it's complex value type)
        # Todo: add new setting `show_title_page=True` to control whether to create the front page or not (and make sure the header creation is aware of this parameter)
        """
            **Params:**
                - ``title`` (str): Title of the report. Placed in the front page.
                - ``subtitle`` (str): Subtitle of the report. Placed in the front page after the title.
                - ``load_fonts`` (list[Dict[str, str]]): Fonts to load for the report.
                 For ``load_fonts`` you need to specify font_family, font_style and font_path as dictionary.

            **Kwargs:**
                - ``orientation`` (str): Orientation of the report, e.g., 'P' for Portrait, 'L' for Landscape.
                - ``unit`` (str): Measurement unit for the report, e.g., 'pt', 'mm', 'cm'.
                - ``format`` (str): Paper format for the report, e.g., 'A4', 'Letter'.
                - ``logo`` (str): Path to the logo file for the report.
                - ``styles`` (Dict): Styles dictionary for additional customization. See below for keys.

            **Styles Dictionary Keys:**

            **Table**
                *title*
                    - ``table.title.font_style`` (str)
                    - ``table.title.font_family`` (str)
                    - ``table.title.font_color`` (str)
                    - ``table.title.font_size`` (float)

                *heading*
                    - ``table.heading.font_style`` (str)
                    - ``table.heading.font_family`` (str)
                    - ``table.heading.font_color`` (str)
                    - ``table.heading.font_size`` (float)
                    - ``table.heading.highlight_color`` (str): color name

                *series*
                    - ``table.series.font_style`` (str)
                    - ``table.series.font_family`` (str)
                    - ``table.series.font_color`` (str)
                    - ``table.series.font_size`` (float)
                    - ``table.series.highlight_color`` (str): color name

                **section**
                    - ``table.section.font_family`` (str)
                    - ``table.section.font_color`` (str)
                    - ``table.section.font_size`` (float)
                    - ``table.section.highlight_color`` (str): color name

            **Chart:**
                **general**
                    - ``chart.highlight_color`` (str)
                    - ``chart.grid_color`` (str)
                    - ``chart.bg_color`` (str): Background color of the chart.
                    - ``chart.chart_border_color`` (str): Border color of the chart. If border is applied
                **legend**
                    - ``chart.legend.border_width`` (int):
                    - ``chart.legend.bg_color`` (str):
                    - ``chart.legend.font_style`` (str):
                    - ``chart.legend.font_family`` (str):
                    - ``chart.legend.font_color`` (str):
                    - ``chart.legend.font_size`` (float):

                **title**
                    - ``chart.title.font_style`` (str)
                    - ``chart.title.font_family`` (str)
                    - ``chart.title.font_color`` (str)
                    - ``chart.title.font_size`` (float)

                **x_axis:**
                    *ticks:*
                        - ``chart.x_axis.ticks.font_size`` (float)
                        - ``chart.x_axis.ticks.font_color`` (str)
                        - ``chart.x_axis.ticks.font_family`` (str)
                    *label:*
                        - ``chart.x_axis.label.font_size`` (float)
                        - ``chart.x_axis.label.font_color`` (str)
                        - ``chart.x_axis.label.font_family`` (str)

                **y_axis:**
                    *ticks:*
                        - ``chart.y_axis.ticks.font_size`` (float)
                        - ``chart.y_axis.ticks.font_color`` (str)
                        - ``chart.y_axis.ticks.font_family`` (str)
                    *label:*
                        - ``chart.y_axis.label.font_size`` (float)
                        - ``chart.y_axis.label.font_color`` (str)
                        - ``chart.y_axis.label.font_family`` (str)

                **series:**
                    - ``chart.series.bar_edge_color`` (str): Border color for bars in bar charts, default black.
                    - ``chart.series.bar_edge_width`` (int): Border width for bars in bar charts.
                    - ``chart.series.bar_face_color`` (str): Fill color for bars in bar charts, default green.
                    - ``chart.series.line_width`` (int): Line width, default 1.
                    - ``chart.series.line_style`` (str): 'solid', 'dot', 'dash', 'longdash', 'dashdot', 'longdashdot'.
                    - ``chart.series.line_color`` (str): Color of the line.
                    - ``chart.series.marker_color`` (str): Color of the marker symbol.
                    - ``chart.series.marker_size`` (int): Size of the marker symbol.
                    - ``chart.series.marker_width`` (int): Marker width default is 0; some, like asterisks, require a width of 1.

            **Report:**
                - ``report.font_style`` (str)
                - ``report.font_family`` (str)
                - ``report.font_color`` (str)
                - ``report.font_size`` (float)
                - ``report.page_margin_left`` (int)
                - ``report.page_margin_top`` (int)

            **Text:**
                - ``text.font_style`` (str): as string: "B", "I", "BI"
                - ``text.font_family`` (str)
                - ``text.font_size`` (float)

            **Chapter:**
                - ``chapter.font_style`` (str)
                - ``chapter.font_family`` (str)
                - ``chapter.font_size`` (float)

            :key kwargs (any): You can provide any key-value pair to customize settings. If Object A contains Object B, settings for Object B can be defined in Object A and will apply to all instances of Object B, unless overridden for a specific instance
        """

        self.CHILDREN = []
        self.title = title
        self.subtitle = subtitle
        self.load_fonts = get_fonts(load_fonts)
        self.abstract = abstract
        self.settings = SettingsContainer(**kwargs)

        if hasattr(self.settings, "styles"):
            self.settings.styles = add_prefix_to_styles("report", self.settings.styles)

        merge_settings(self, self.settings.__dict__)

    def add(self, report_child : Union[Grid, Table, Text, PageBreak, Chart, Chapter] = None):
        if report_child is not None and isinstance(report_child, (Grid, Table, Text, PageBreak, Chart, Chapter)):
            copy_report_child = copy.deepcopy(report_child)
            self.CHILDREN.append(copy_report_child)
            merge_settings(self, self.settings.__dict__)
        else: raise Exception("Report: report child of wrong type or None")

    # TODO: Output needs to know about mapper, so regarding circularity, extracting output would help us.
    # Should be part of report mapper
    def output(self, file_name: str = "report", cleanup: bool = True):
        """
            :args file_name (str): name of the report output file
            :args cleanup (bool): if True, figures will be kept within /tmp directory
        """

        pdf = PDF(
            orientation = self.settings.orientation,
            unit = self.settings.unit,
            format = self.settings.format,
            report_styles=self.settings.styles["report"],
            report_title=self.title,
        )

        add_custom_fonts(pdf, self.load_fonts)

        pdf.set_margins(left=self.settings.styles["report"]["page_margin_left"],
                         top=self.settings.styles["report"]["page_margin_top"])

        reportMapper = ReportMapper(pdf)

        reportMapper.compile(self, file_name, cleanup)

    def set_front_page(self, title: str = "", subtitle: str = "", abstract: str="", logo: Optional[str] = None):
        if bool(title):
            self.title = title
        if bool(subtitle):
            self.subtitle = subtitle
        if bool(abstract):
            self.abstract = abstract
        if bool(logo):
            self.settings.logo = logo






from fpdf import enums

# Todo add some space after line, between line and first footnote

def add_footnotes_at_bottom(pdf, footnotes, reference_nr, last_y = None, **kwargs):

    font_size = kwargs.get('font_size')
    font_family = kwargs.get('font_family')
    font_style = kwargs.get('font_style')
    footer_padding = kwargs.get('footer_padding')
    footer_top_margin = kwargs.get('footer_top_margin')


    bottom_padding = 10
    b_margin = pdf.b_margin

    # prevent auto page break - this is a MUST!!!!!
    # else it would push content and footnotes to new page if slightly overflows
    pdf.set_auto_page_break(auto=False)

    pdf.set_font(font_family, font_style, font_size)

    #Sets line color above the footnotes
    pdf.set_draw_color(0, 0, 0)

    # current b_margin = page_margin_bottom + footer_height + footer_top_margin
    footer_position = pdf.h - b_margin + footer_top_margin

    if last_y is None:
        footnote_y_position = pdf.h - b_margin + footer_padding + footer_top_margin
        pdf.line(x1=pdf.l_margin, y1=footer_position, x2=pdf.w - pdf.r_margin, y2=footer_position)
    else:
        footnote_y_position = last_y

    pdf.set_y(footnote_y_position)

    for ref, footnote in zip(reference_nr, footnotes):
        formatted_footnote = f"[{ref}] {footnote}"
        pdf.multi_cell(w=pdf.w - pdf.l_margin - pdf.r_margin, h=None, text=formatted_footnote, align='L', new_x=enums.XPos.LEFT)
        last_y = pdf.get_y() + bottom_padding

    # Reset auto page break and b_margin
    # Todo: drop footer height, rename footer padding to bottom_margin_padding, and footnotes pos y would be pdf.h - b.margin
    pdf.set_auto_page_break(auto=True, margin = b_margin)

    return last_y



import fitz  # PyMuPDF
import gc

def overlay_pdfs(base_pdf_path, pdf_data, output_pdf_path):
    # 'with' handles the base PDF file handle
    with fitz.open(base_pdf_path) as base_pdf:

        for i, pdf_info in enumerate(pdf_data):
            page_num = pdf_info['page'] - 1

            # 1. Read the file into a memory buffer
            # 2. Open the PDF FROM that buffer
            # Using 'with' here is CRITICAL to release the memory buffer
            # after the page is placed.
            with open(pdf_info['pdf_path'], "rb") as f:
                chart_buffer = f.read()

            with fitz.open(stream=chart_buffer, filetype="pdf") as overlay_pdf:
                base_page = base_pdf[page_num]

                position = fitz.Rect(
                    pdf_info['x'],
                    pdf_info['y'],
                    pdf_info['x'] + pdf_info['width'],
                    pdf_info['y'] + pdf_info['height']
                )

                # Place the overlay PDF
                base_page.show_pdf_page(position, overlay_pdf, 0)

            # --- GARBAGE COLLECTION ---
            # Every 100 charts, we force a deep clean
            if (i + 1) % 100 == 0:
                # Explicitly delete the buffer variable to help GC
                del chart_buffer

                # Shrink the MuPDF internal caches (images/fonts)
                fitz.TOOLS.store_shrink(100)

                # Force Python to collect unreachable memory cycles
                gc.collect()

        base_pdf.save(output_pdf_path, garbage=1, deflate=True)

    print(f"Overlay PDF saved: {output_pdf_path}")


"""
Prepare shared (base-level) trace update settings for Plotly plots.

This function returns general trace configuration that applies uniformly
across all traces — where no fine-grained (per-trace) control is needed
between multivariate and single-series cases.

"""
def update_base_traces(settings):
    yaxis = getattr(settings, "yaxis", "left")
    if settings.series_type == "line":
         update_traces = {
            "mode": settings.markers_mode,
            "marker": {
                "symbol": settings.marker_symbol,
                "size": settings.styles["chart"]["series"]["marker_size"],
                "line.width": settings.styles["chart"]["series"]["marker_width"],
                "line.color": settings.styles["chart"]["series"]["marker_color"] or settings.styles["chart"]["series"]["line_color"],
                "color": settings.styles["chart"]["series"]["marker_color"] or settings.styles["chart"]["series"]["line_color"],
            },
            "yaxis": "y2" if yaxis == "right" else "y1"
         }
    elif "bar" in settings.series_type.lower():
         update_traces = {
            "type": "bar",
            "marker": {
                "line.color": settings.styles["chart"]["series"]["bar_edge_color"],
                "line.width": settings.styles["chart"]["series"]["bar_edge_width"],
            },
              "yaxis": "y2" if yaxis == "right" else "y1"
        }
    else: update_traces = {}

    return update_traces

def apply_per_trace_styles(
    fig,
    line_colors=None,
    line_widths=None,
    line_styles=None,
    bar_colors=None,
    num_variants=None,
):
    """
    Apply per-trace styling to a Plotly figure.

    Supports passing a single value or a list for each style type.
    """

    if not fig.data:
        return

    traces = fig.data[-num_variants:]

    for idx, trace in enumerate(traces):

        if not trace.name:
            trace.showlegend = False

        if trace.type == 'scatter':

            color = get_value(line_colors, idx)
            if color:
                trace.line.color = color
                # Fix: Apply the specific trace color to the markers as well
                if trace.marker:
                    trace.marker.color = color
                    trace.marker.line.color = color

            width = get_value(line_widths, idx)
            if width:
                trace.line.width = width

            dash = get_value(line_styles, idx)
            if dash:
                trace.line.dash = dash

        elif trace.type == 'bar':
            color = get_value(bar_colors, idx)
            if color:
                trace.marker.color = color


def get_value(value, i):
    """
    Function to retrieve the appropriate style value for the n-th trace.
    Mainly used for apply_per_trace_styles()

    :param value: Either a single value or a list of values.
    :param i: trace index
    :return: value (str)
    """
    if isinstance(value, list):
        # If a list, return value based on trace index, and cycle
        return value[i % len(value)]
    else: return value  # bcs single value can also be passed



import copy

from rephorm.object_mappers.table_section_mapper import TableSectionMapper
from rephorm.objects.table_series import TableSeries
from rephorm.objects._utilities.settings_container import SettingsContainer
from rephorm.decorators.settings_validation import validate_kwargs
from rephorm.utility.add_style_prefix import add_prefix_to_styles


class TableSection:

    @validate_kwargs
    def __init__(self, title: str = None, **kwargs):
        """
            :param title: (str) Title of the table section.
            :key styles (Dict): Styles dictionary for the table section (refer to documentation).
            :key highlight (dp.Span): Span of the highlight in the table section. (applies to all table_series added within the section)
            :key show_units (bool): To show/hide units for the table section. (applies to all table_series added within the section)
        """
        self.TITLE = title
        self.CHILDREN = []
        self.settings = SettingsContainer(**kwargs)
        if hasattr(self.settings, "styles"):
            self.settings.styles = add_prefix_to_styles("table.section", self.settings.styles)

    def add (self, table_section_child: TableSeries = None):

        self._on_add_callback()

        if table_section_child is not None and isinstance(table_section_child, TableSeries):
            copy_table_section_child = copy.deepcopy(table_section_child)
            self.CHILDREN.append(copy_table_section_child)
        else:
            raise Exception("TableSection: Table Section child of wrong type or None")

    def _get_mapper(self):
        return TableSectionMapper(self)

    def __repr__(self):
        return f"{type(self).__name__}"

    def _on_add_callback(self):
        if hasattr(self.settings, 'styles'):
            highlight_color = self.settings.styles.get("table.section.highlight_color")
            if highlight_color:
                self.settings.styles.setdefault("table.series.highlight_color", highlight_color)
                self.settings.styles.setdefault("table.section.highlight_color", highlight_color)




from fpdf import FontFace

from rephorm.object_mappers._utilities.table_mapper_utility import (
    get_max_numerical_width,
    get_max_width,
)
from rephorm.utility.report.add_footnotes import add_footnotes_at_bottom
from rephorm.utility.report.footnotes_counter import generate_footnote_numbers
from rephorm.utility.report.report_utility import get_page_dimensions
from rephorm.utility.report.resolve_color import resolve_color
from rephorm.utility.report.title_utility import add_title_with_references


class TableMapper:
    def __init__(self, table):
        self.table = table

    def add_to_pdf(self, **kwargs):

        pdf = kwargs["pdf"]
        highlight = []
        x = kwargs["x"]
        w = kwargs["w"]
        h = kwargs["h"]
        y = kwargs["y"]

        title_bottom_padding = self.table.settings.styles["report"]["title_bottom_padding"]
        #Todo: Consider extracting context based on caller. So we do not need to deal with string checks. Discuss with Sergey.
        context = kwargs.get("context", None)
        custom_width = self.table.width
        footer_top_padding = self.table.settings.styles["report"]["footer"]["top_padding"]
        footer_top_margin = self.table.settings.styles["report"]["footer"]["top_margin"]
        footer_last_y = kwargs["footer_last_y"] if "footer_last_y" in kwargs else None

        # Todo: Consider replacing (full_width_grid) conditional based on grid row settings? if nrow and ncol = 1? Talk to Sergey why
        full_width_grid = (w == pdf.epw)
        is_multi_cell_grid = context == "grid" and not full_width_grid

        skip_cols = 1
        headers = [""]

        pdf.set_xy(x, y)
        page_width, _ = get_page_dimensions(pdf)

        num_col_width = get_max_numerical_width(
            self.table,
            pdf,
            font_family=self.table.settings.styles["table"]["series"]["font_family"],
            font_size=self.table.settings.styles["table"]["series"]["font_size"],
            biggest_nr="10000",
        )

        title_col_width = get_max_width(
            self.table,
            pdf,
            font_family=self.table.settings.styles["table"]["series"]["font_family"],
            font_size=self.table.settings.styles["table"]["series"]["font_size"],
            parameter="title",
            max_width=0,
        )
        unit_col_width = get_max_width(
            self.table,
            pdf,
            font_family=self.table.settings.styles["table"]["series"]["font_family"],
            font_size=self.table.settings.styles["table"]["series"]["font_size"],
            parameter="unit",
            max_width=0,
        )

        if self.table.settings.highlight is not None:
            highlight = self.table._get_table_highlight()

        span = self.table._get_span()

        if self.table.settings.show_units:
            skip_cols += 1

        if self.table.settings.show_units:
            headers = headers + ["Units"]

        heading_data = [
            headers + list(span.to_compact_strings()),
        ]

        heading = heading_data[0][skip_cols:]

        num_col_padding = 10

        col_widths = (
            [
                title_col_width + 10
            ]  # TODO: first column width | We introduce param for this. the name of param: title_col_width
            + [unit_col_width + 0]
            * self.table.settings.show_units  # TODO: second column width | We introduce param for this. the name of param: unit_col_width
            + [num_col_width + num_col_padding] * len(heading)  # numerical columns width
        )

        # we use effective page width here (excluding margins)
        page_width = pdf.epw

        natural_table_width = sum(col_widths) + 10

        if natural_table_width > page_width:
            table_width = page_width
        else:
            if is_multi_cell_grid:
                table_width = w
            else:
                table_width = natural_table_width

        text_align = ["L"] * skip_cols + ["R"] * len(heading)

        h_data = headers + heading

        # ----------------- Set Table Title -----------------

        footnote_references = generate_footnote_numbers(self.table.footnotes)
        y = add_title_with_references(
            pdf=pdf,
            title=self.table.title,
            references=footnote_references,
            font_family=self.table.settings.styles["table"]["title"]["font_family"],
            font_size=self.table.settings.styles["table"]["title"]["font_size"],
            font_style=self.table.settings.styles["table"]["title"]["font_style"],
            title_y=y,
            cell_width=w,
        )
        if bool(self.table.footnotes):
            footer_last_y = add_footnotes_at_bottom(
                pdf,
                self.table.footnotes,
                footnote_references,
                last_y=footer_last_y,
                font_family=self.table.settings.styles["footnotes"]["font_family"],
                font_size=self.table.settings.styles["footnotes"]["font_size"],
                font_style=self.table.settings.styles["footnotes"]["font_style"],
                footer_padding=footer_top_padding,
                footer_top_margin=footer_top_margin,
            )

        # Todo: Consider: title padding could be part of add_title_with_ref function, so we only pass here Y after calling that func)
        pdf.set_y(y + title_bottom_padding)

        # ----------------- Set Table Header -----------------
        with pdf.table(
            width=custom_width if custom_width else table_width,
            col_widths=col_widths,
            text_align=text_align,
            align="l" if is_multi_cell_grid else "c",
            # TODO: ADD border settings to styles for table
            borders_layout="SINGLE_TOP_LINE",
            # line height sets the height of rows, 1,4 is multiplier. (We cannot precisely extract height of the text)
            # Todo: Consider extracting this to param/styles, like inrow_offset...
            line_height=self.table.settings.styles["table"]["series"]["font_size"]
            * 1.4,
        ) as pdf_table:

            heading_row = pdf_table.row()
            header_color = resolve_color(
                self.table.settings.styles["table"]["heading"]["highlight_color"]
            )

            for i, header in enumerate(h_data):
                # Todo: Remove Grey and etc params from table object and move them to styles
                style = FontFace(
                    emphasis=self.table.settings.styles["table"]["heading"][
                        "font_style"
                    ],
                    size_pt=self.table.settings.styles["table"]["heading"]["font_size"],
                    fill_color=header_color if i in highlight else "#FFFFFF00",
                    family=self.table.settings.styles["table"]["heading"][
                        "font_family"
                    ],
                    color=resolve_color(
                        self.table.settings.styles["table"]["heading"]["font_color"]
                    ),
                )

                heading_row.cell(header, style=style)

            # ----------------- Set Table Rows -----------------
            for item in self.table.CHILDREN:
                mapper = item._get_mapper()
                mapper.add_to_pdf(
                    span=span,
                    pdf_table=pdf_table,
                    highlight=highlight,
                    n_cols=len(col_widths),
                    show_units=self.table.settings.show_units,
                    footer_last_y=footer_last_y,
                )



# The 'prefixes' tuple contains the current object name prefixes.
# When defining styles, avoid using object names as keys.
# For example, a structure like Grid: { chart: { } }
# will cause issues, because chart is also an object name.

prefixes = (
    "chart.",
    "grid.",
    "report.",
    "chapter.",
    "table.",
    "text.",
)

def flatten_dict(d, parent_key='', sep='.'):
    items = []
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)

def add_prefix_to_styles(obj_name, styles):
    flatten_styles = flatten_dict(styles)
    return {
        key if key.startswith(prefixes) else f"{obj_name}.{key}": value
        for key, value in flatten_styles.items()
    }



import copy
from typing import Optional, List, Union

from rephorm.objects.table_section import TableSection
from rephorm.objects.table_series import TableSeries
from rephorm.object_mappers.table_mapper import TableMapper
from rephorm.objects._utilities.settings_container import SettingsContainer
from rephorm.utility.add_style_prefix import add_prefix_to_styles
from rephorm.utility.report.range_utility import get_span, get_highlight
from rephorm.decorators.settings_validation import validate_kwargs
from rephorm.utility.report.table_utility import get_table_highlight


class Table:

    @validate_kwargs
    def __init__(self, title: str = "", footnotes: Optional[List[str]] = None, width: int = None, **kwargs):
        """
            :param title: Title of the table.
            :param footnotes: Footnotes for the title.
            :param width (int): The maximum width allowed for rendering the table.
            :key span (dp.Span, Required): Span of the table.
            :key highlight (dp.Span): Span of the highlight.
            :key show_units (bool): Display/hide units in the table.
            :key frequency (str): Frequency of the data in the table.
            :key decimal_precision (int): Sets number of decimal places after the decimal point.
            :key compare_style (str): Comparison style for the table series. "diff", "pct".
            :key layout (Dict): Layout configuration for the table (Plotly specific).
            :key styles (Dict): Styles dictionary for additional customization (for details refer to report object).
        """
        self.CHILDREN = []
        self.footnotes = footnotes
        self.title = title
        self.width = width
        self.settings = SettingsContainer(**kwargs)

        if hasattr(self.settings, "styles"):
            self.settings.styles = add_prefix_to_styles("table", self.settings.styles)


    def add(self, table_child: Union[TableSeries, TableSection] = None,):

        self._on_add_callback()

        if table_child is not None and isinstance(table_child, (TableSeries, TableSection)):
            copy_table_child = copy.deepcopy(table_child)
            self.CHILDREN.append(copy_table_child)
        else:
            raise Exception("Table: Table child of wrong type or None")

    def _get_mapper(self):
        return TableMapper(self)

    def _get_span(self):
        return get_span(self)

    def _get_highlight(self):
        return get_highlight(self)

    def _get_table_highlight(self):
        return get_table_highlight(self)

    # Because sometimes we may want to print the name of the object
    def __repr__(self):
        return f"{type(self).__name__}"

    def _on_add_callback(self):

        if hasattr(self.settings, 'styles'):
            highlight_color = self.settings.styles.get("table.highlight_color")

            if highlight_color:
                # Set default values for various highlight_color properties
                self.settings.styles.setdefault("table.series.highlight_color", highlight_color)
                self.settings.styles.setdefault("table.heading.highlight_color", highlight_color)
                self.settings.styles.setdefault("table.section.highlight_color", highlight_color)
